"""JSON parser for Opta F9 feeds."""
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from ...base import MissingDataError
from .base import OptaJSONParser, assertget


class F9JSONParser(OptaJSONParser):
    """Extract data from a Opta F9 data stream.

    Parameters
    ----------
    path : str
        Path of the data file.
    """

    def _get_feed(self) -> Dict[str, Any]:
        for node in self.root:
            if "OptaFeed" in node["data"].keys():
                return node
        raise MissingDataError

    def _get_doc(self) -> Dict[str, Any]:
        f9 = self._get_feed()
        data = assertget(f9, "data")
        optafeed = assertget(data, "OptaFeed")
        optadocument = assertget(optafeed, "OptaDocument")[0]
        return optadocument

    def _get_stats(self, obj: Dict[str, Any]) -> Dict[str, Any]:
        if "Stat" not in obj:
            return {}

        stats = {}
        statobj = obj["Stat"] if isinstance(obj["Stat"], list) else [obj["Stat"]]
        for stat in statobj:
            stats[stat["@attributes"]['Type']] = stat["@value"]
        return stats

    def _get_name(self, obj: Dict[str, Any]) -> Optional[str]:
        if "Known" in obj and obj["Known"].strip():
            return obj["Known"]
        if "First" in obj and "Last" in obj and obj["Last"].strip() or obj["First"].strip():
            return (obj["First"] + " " + obj["Last"]).strip()
        return None

    def extract_games(self) -> Dict[int, Dict[str, Any]]:
        """Return a dictionary with all available games.

        Returns
        -------
        dict
            A mapping between game IDs and the information available about
            each game in the data stream.
        """
        optadocument = self._get_doc()
        attr = assertget(optadocument, "@attributes")
        matchdata = assertget(optadocument, "MatchData")
        competition = assertget(optadocument, "Competition")
        competitionstat = self._get_stats(competition)
        venue = assertget(optadocument, "Venue")
        matchofficial = assertget(matchdata, "MatchOfficial")
        matchinfo = assertget(matchdata, "MatchInfo")
        matchstat = self._get_stats(matchdata)
        teamdata = assertget(matchdata, "TeamData")
        scores = {}
        for t in teamdata:
            scores[t["@attributes"]["Side"]] = t["@attributes"]["Score"]

        game_id = int(assertget(attr, "uID")[1:])
        game_dict = dict(
            # Fields required by the base schema
            game_id=game_id,
            competition_id=int(assertget(assertget(competition, "@attributes"), "uID")[1:]),
            season_id=assertget(competitionstat, "season_id"),
            game_day=competitionstat["matchday"] if "matchday" in competitionstat else None,
            game_date=datetime.strptime(assertget(matchinfo, "Date"), "%Y%m%dT%H%M%S%z").replace(
                tzinfo=None
            ),
            # home_team_id=see below
            # away_team_id=see below
            # Optional fields
            home_score=int(scores["Home"]),
            away_score=int(scores["Away"]),
            duration=int(assertget(matchstat, "match_time")),
            referee=self._get_name(matchofficial["OfficialName"])
            if "OfficialName" in matchofficial
            else None,
            venue=venue["Name"] if "Name" in venue else None,
            attendance=int(matchinfo["Attendance"]) if "Attendance" in matchinfo else None,
            # home_manager=see below
            # away_manager=see below
        )
        for team in teamdata:
            teamattr = assertget(team, '@attributes')
            side = assertget(teamattr, 'Side')
            teamid = assertget(teamattr, 'TeamRef')
            score = assertget(teamattr, 'Score')
            manager = (
                self._get_name(team["TeamOfficial"]["PersonName"])
                if "TeamOfficial" in team
                else None
            )
            if side == 'Home':
                game_dict['home_team_id'] = int(teamid[1:])
                game_dict['home_score'] = int(score)
                game_dict['home_manager'] = manager
            else:
                game_dict['away_team_id'] = int(teamid[1:])
                game_dict['away_score'] = int(score)
                game_dict['away_manager'] = manager
        return {game_id: game_dict}

    def extract_teams(self) -> Dict[int, Dict[str, Any]]:
        """Return a dictionary with all available teams.

        Returns
        -------
        dict
            A mapping between team IDs and the information available about
            each team in the data stream.
        """
        optadocument = self._get_doc()
        root_teams = assertget(optadocument, "Team")

        teams = {}
        for team in root_teams:
            if "id" in team.keys():
                nameobj = team.get("nameObj")
                team_id = int(team["id"])
                teams[team_id] = dict(
                    # Fields required by the base schema
                    team_id=team_id,
                    team_name=nameobj.get("name"),
                )
        return teams

    def extract_players(self) -> Dict[Tuple[int, int], Dict[str, Any]]:
        """Return a dictionary with all available players.

        Returns
        -------
        dict
            A mapping between (game ID, player ID) tuples and the information
            available about each player in the data stream.
        """
        optadocument = self._get_doc()
        attr = assertget(optadocument, "@attributes")
        game_id = int(assertget(attr, "uID")[1:])
        root_teams = assertget(optadocument, "Team")
        lineups = self.extract_lineups()

        players = {}
        for team in root_teams:
            team_id = int(team["@attributes"]["uID"].replace("t", ""))
            for player in team["Player"]:
                player_id = int(player["@attributes"]["uID"].replace("p", ""))

                assert "nameObj" in player["PersonName"]
                nameobj = player["PersonName"]["nameObj"]
                if not nameobj.get("is_unknown"):
                    player = dict(
                        # Fields required by the base schema
                        game_id=game_id,
                        team_id=team_id,
                        player_id=player_id,
                        player_name=self._get_name(player["PersonName"]),
                        # is_starter=
                        # minutes_played=
                        # jersey_number=
                        # Fields required by the opta schema
                        # starting_position=
                        # Optional fields
                        # height="?",
                        # weight="?",
                        # age="?",
                    )
                    if player_id in lineups[team_id]["players"]:
                        player = dict(
                            **player,
                            jersey_number=lineups[team_id]["players"][player_id]["jersey_number"],
                            starting_position=lineups[team_id]["players"][player_id][
                                "starting_position_name"
                            ],
                            is_starter=lineups[team_id]["players"][player_id]["is_starter"],
                            minutes_played=lineups[team_id]["players"][player_id][
                                "minutes_played"
                            ],
                        )
                    players[(game_id, player_id)] = player
        return players

    def extract_lineups(self) -> Dict[int, Dict[str, Any]]:
        """Return a dictionary with the lineup of each team.

        Raises
        ------
        MissingDataError
            If teams data is not available in the stream.

        Returns
        -------
        dict
            A mapping between team IDs and the information available about
            each team's lineup in the data stream.
        """
        optadocument = self._get_doc()
        attr = assertget(optadocument, "@attributes")

        try:
            rootf9 = optadocument["MatchData"]["TeamData"]
        except KeyError as e:
            raise MissingDataError from e
        matchstats = optadocument["MatchData"]["Stat"]
        matchstats = [matchstats] if isinstance(matchstats, dict) else matchstats
        matchstatsdict = {stat["@attributes"]["Type"]: stat["@value"] for stat in matchstats}

        lineups: Dict[int, Dict[str, Any]] = {}
        for team in rootf9:
            # lineup attributes
            team_id = int(team["@attributes"]["TeamRef"].replace("t", ""))
            lineups[team_id] = dict(players=dict())
            # substitutes
            subst = [s["@attributes"] for s in team["Substitution"]]
            # red cards
            red_cards = {
                int(e["@attributes"]["PlayerRef"].replace("p", "")): e["@attributes"]["Time"]
                for e in team.get("Booking", [])
                if "CardType" in e["@attributes"]
                and e["@attributes"]["CardType"] in ["Red", "SecondYellow"]
                and "PlayerRef" in e["@attributes"]  # not defined if a coach receives a red card
            }
            for player in team["PlayerLineUp"]["MatchPlayer"]:
                attr = player["@attributes"]
                player_id = int(attr["PlayerRef"].replace("p", ""))
                playerstatsdict = {
                    stat["@attributes"]["Type"]: stat["@value"] for stat in player["Stat"]
                }
                sub_on = next(
                    (
                        item["Time"]
                        for item in subst
                        if "Retired" not in item and item["SubOn"] == f"p{player_id}"
                    ),
                    matchstatsdict["match_time"] if attr["Status"] == "Sub" else 0,
                )
                sub_off = next(
                    (item["Time"] for item in subst if item["SubOff"] == f"p{player_id}"),
                    matchstatsdict["match_time"]
                    if player_id not in red_cards
                    else red_cards[player_id],
                )
                minutes_played = sub_off - sub_on
                lineups[team_id]["players"][player_id] = dict(
                    jersey_number=attr["ShirtNumber"],
                    starting_position_name=attr["Position"],
                    starting_position_id=attr["position_id"],
                    is_starter=attr["Status"] == "Start",
                    minutes_played=minutes_played,
                    **playerstatsdict,
                )
        return lineups

    def extract_teamgamestats(self) -> List[Dict[str, Any]]:
        """Return some aggregated statistics of each team.

        Raises
        ------
        MissingDataError
            If teams data is not available in the stream.

        Returns
        -------
        list(dict)
            A dictionary with aggregated team statistics for each team.
        """
        optadocument = self._get_doc()
        attr = assertget(optadocument, "@attributes")
        game_id = int(assertget(attr, "uID")[1:])

        try:
            rootf9 = optadocument["MatchData"]["TeamData"]
        except KeyError as e:
            raise MissingDataError from e
        teams_gamestats = []
        for team in rootf9:
            attr = team["@attributes"]
            statsdict = self._get_stats(team)

            team_gamestats = dict(
                game_id=game_id,
                team_id=int(attr["TeamRef"].replace("t", "")),
                side=attr["Side"],
                score=attr["Score"],
                shootout_score=attr["ShootOutScore"],
                **statsdict,
            )

            teams_gamestats.append(team_gamestats)
        return teams_gamestats



"""
SoccerAction
~~~~~~~~~~~~

SoccerAction is a Python package for objectively quantifying the impact of the
individual actions performed by soccer players using event stream data.

Full documentation is at <https://ml-kuleuven.github.io/socceraction/>.
:copyright: (c) 2020 by DTAI KU Leuven.
:license: MIT, see LICENSE for more details.
"""
__version__ = '1.4.2'



"""Implementation of the SPADL language."""

__all__ = [
    'opta',
    'statsbomb',
    'wyscout',
    'config',
    'SPADLSchema',
    'bodyparts_df',
    'actiontypes_df',
    'results_df',
    'add_names',
    'play_left_to_right',
]

from . import config, opta, statsbomb, wyscout
from .config import actiontypes_df, bodyparts_df, results_df
from .schema import SPADLSchema
from .utils import add_names, play_left_to_right



"""SPADL schema for Opta data."""
from typing import Optional

import pandas as pd
import pandera as pa
from pandera.typing import DateTime, Object, Series

from socceraction.data.schema import (
    CompetitionSchema,
    EventSchema,
    GameSchema,
    PlayerSchema,
    TeamSchema,
)


class OptaCompetitionSchema(CompetitionSchema):
    """Definition of a dataframe containing a list of competitions and seasons."""


class OptaGameSchema(GameSchema):
    """Definition of a dataframe containing a list of games."""

    home_score: Optional[Series[int]] = pa.Field(nullable=True)
    """The final score of the home team."""
    away_score: Optional[Series[int]] = pa.Field(nullable=True)
    """The final score of the away team."""
    duration: Optional[Series[int]] = pa.Field(nullable=True)
    """The total duration of the game in minutes."""
    referee: Optional[Series[str]] = pa.Field(nullable=True)
    """The name of the referee."""
    venue: Optional[Series[str]] = pa.Field(nullable=True)
    """The name of the stadium where the game was played."""
    attendance: Optional[Series[int]] = pa.Field(nullable=True)
    """The number of people who attended the game."""
    home_manager: Optional[Series[str]] = pa.Field(nullable=True)
    """The name of the manager of the home team."""
    away_manager: Optional[Series[str]] = pa.Field(nullable=True)
    """The name of the manager of the away team."""


class OptaPlayerSchema(PlayerSchema):
    """Definition of a dataframe containing the list of players of a game."""

    starting_position: Series[str]
    """The starting position of the player."""


class OptaTeamSchema(TeamSchema):
    """Definition of a dataframe containing the list of teams of a game."""


class OptaEventSchema(EventSchema):
    """Definition of a dataframe containing event stream data of a game."""

    timestamp: Series[DateTime]
    """Time in the match the event takes place, recorded to the millisecond."""
    minute: Series[int]
    """The minutes on the clock at the time of this event."""
    second: Series[int] = pa.Field(ge=0, le=59)
    """The second part of the timestamp."""
    outcome: Series[bool]
    """Whether the event had a successful outcome or not."""
    start_x: Series[float] = pa.Field(nullable=True)
    """The x coordinate of the location where the event started."""
    start_y: Series[float] = pa.Field(nullable=True)
    """The y coordinate of the location where the event started."""
    end_x: Series[float] = pa.Field(nullable=True)
    """The x coordinate of the location where the event ended."""
    end_y: Series[float] = pa.Field(nullable=True)
    """The y coordinate of the location where the event ended."""
    qualifiers: Series[Object]
    """A JSON object containing the Opta qualifiers of the event."""
    assist: Optional[Series[bool]]
    """Whether the event was an assist or not."""
    keypass: Optional[Series[bool]]
    """Whether the event was a keypass or not."""
    goal: Optional[Series[bool]]
    """Whether the event was a goal or not."""
    shot: Optional[Series[bool]]
    """Whether the event was a shot or not."""
    touch: Optional[Series[bool]]
    """Whether the event was a on-the-ball action or not."""
    related_player_id: Optional[Series[pd.Int64Dtype]] = pa.Field(nullable=True)
    """The ID of a second player that was involved in this event."""



"""JSON parser for Stats Perform MA3 feeds."""
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd

from ...base import MissingDataError
from .base import OptaJSONParser, _get_end_x, _get_end_y, assertget


class MA3JSONParser(OptaJSONParser):
    """Extract data from a Stats Perform MA3 data stream.

    Parameters
    ----------
    path : str
        Path of the data file.
    """

    _position_map = {
        1: "Goalkeeper",
        2: "Defender",
        3: "Midfielder",
        4: "Forward",
        5: "Substitute",
    }

    def _get_match_info(self) -> Dict[str, Any]:
        if "matchInfo" in self.root:
            return self.root["matchInfo"]
        raise MissingDataError

    def _get_live_data(self) -> Dict[str, Any]:
        if "liveData" in self.root:
            return self.root["liveData"]
        raise MissingDataError

    def extract_competitions(self) -> Dict[Tuple[str, str], Dict[str, Any]]:
        """Return a dictionary with all available competitions.

        Returns
        -------
        dict
            A mapping between competion IDs and the information available about
            each competition in the data stream.
        """
        match_info = self._get_match_info()
        season = assertget(match_info, "tournamentCalendar")
        competition = assertget(match_info, "competition")
        competition_id = assertget(competition, "id")
        season_id = assertget(season, "id")
        season = dict(
            # Fields required by the base schema
            season_id=season_id,
            season_name=assertget(season, "name"),
            competition_id=competition_id,
            competition_name=assertget(competition, "name"),
        )
        return {(competition_id, season_id): season}

    def extract_games(self) -> Dict[str, Dict[str, Any]]:
        """Return a dictionary with all available games.

        Returns
        -------
        dict
            A mapping between game IDs and the information available about
            each game in the data stream.
        """
        match_info = self._get_match_info()
        game_id = assertget(match_info, "id")
        season = assertget(match_info, "tournamentCalendar")
        competition = assertget(match_info, "competition")
        contestant = assertget(match_info, "contestant")
        game_date = assertget(match_info, "date")[0:10]
        game_time = assertget(match_info, "time")[0:8]
        game_datetime = f"{game_date}T{game_time}"
        venue = assertget(match_info, "venue")
        game_obj = dict(
            game_id=game_id,
            competition_id=assertget(competition, "id"),
            season_id=assertget(season, "id"),
            game_day=int(match_info["week"]) if "week" in match_info else None,
            game_date=datetime.strptime(game_datetime, "%Y-%m-%dT%H:%M:%S"),
            home_team_id=self._extract_team_id(contestant, "home"),
            away_team_id=self._extract_team_id(contestant, "away"),
            venue=assertget(venue, "shortName"),
        )
        live_data = self._get_live_data()
        if "matchDetails" in live_data:
            match_details = assertget(live_data, "matchDetails")
            if "matchLengthMin" in match_details:
                game_obj["duration"] = assertget(match_details, "matchLengthMin")
            if "scores" in match_details:
                scores = assertget(match_details, "scores")
                game_obj["home_score"] = assertget(scores, "total")["home"]
                game_obj["away_score"] = assertget(scores, "total")["away"]

        return {game_id: game_obj}

    def extract_teams(self) -> Dict[str, Dict[str, Any]]:
        """Return a dictionary with all available teams.

        Returns
        -------
        dict
            A mapping between team IDs and the information available about
            each team in the data stream.
        """
        match_info = self._get_match_info()
        contestants = assertget(match_info, "contestant")
        teams = {}
        for contestant in contestants:
            team_id = assertget(contestant, "id")
            team = dict(
                # Fields required by the base schema
                team_id=team_id,
                team_name=assertget(contestant, "name"),
            )
            teams[team_id] = team
        return teams

    def extract_players(self) -> Dict[Tuple[str, str], Dict[str, Any]]:  # noqa: C901
        """Return a dictionary with all available players.

        Returns
        -------
        dict
            A mapping between (game ID, player ID) tuples and the information
            available about each player in the data stream.
        """
        match_info = self._get_match_info()
        game_id = assertget(match_info, "id")
        live_data = self._get_live_data()
        events = assertget(live_data, "event")

        game_duration = self._extract_duration()
        playerid_to_name = {}

        players_data: Dict[str, List[Any]] = {
            "starting_position_id": [],
            "player_id": [],
            "team_id": [],
            "position_in_formation": [],
            "jersey_number": [],
        }
        red_cards = {}

        for event in events:
            event_type = assertget(event, "typeId")
            if event_type == 34:
                team_id = assertget(event, "contestantId")
                qualifiers = assertget(event, "qualifier")
                for q in qualifiers:
                    qualifier_id = assertget(q, "qualifierId")
                    value = assertget(q, "value")
                    value = value.split(", ")
                    if qualifier_id == 30:
                        players_data["player_id"] += value
                        team = [team_id for _ in range(len(value))]
                        players_data["team_id"] += team
                    elif qualifier_id == 44:
                        value = [int(v) for v in value]
                        players_data["starting_position_id"] += value
                    elif qualifier_id == 131:
                        value = [int(v) for v in value]
                        players_data["position_in_formation"] += value
                    elif qualifier_id == 59:
                        value = [int(v) for v in value]
                        players_data["jersey_number"] += value
            elif event_type == 17 and "playerId" in event:
                qualifiers = assertget(event, "qualifier")
                for q in qualifiers:
                    qualifier_id = assertget(q, "qualifierId")
                    if qualifier_id in [32, 33]:
                        red_cards[event["playerId"]] = event["timeMin"]

            player_id = event.get("playerId")
            if player_id is None:
                continue
            player_name = assertget(event, "playerName")
            if player_id not in playerid_to_name:
                playerid_to_name[player_id] = player_name

        df_players_data = pd.DataFrame.from_dict(players_data)  # type: ignore

        substitutions = list(self.extract_substitutions().values())
        substitutions_columns = ["player_id", "team_id", "minute_start", "minute_end"]
        df_substitutions = pd.DataFrame(substitutions, columns=substitutions_columns)
        df_substitutions = df_substitutions.groupby(["player_id", "team_id"]).max().reset_index()
        df_substitutions["minute_start"] = df_substitutions["minute_start"].fillna(0)
        df_substitutions["minute_end"] = df_substitutions["minute_end"].fillna(game_duration)

        if df_substitutions.empty:
            df_players_data["minute_start"] = 0
            df_players_data["minute_end"] = game_duration
        else:
            df_players_data = df_players_data.merge(
                df_substitutions, on=["team_id", "player_id"], how="left"
            )
        df_players_data["minute_end"] = df_players_data.apply(
            lambda row: red_cards[row["player_id"]]
            if row["player_id"] in red_cards
            else row["minute_end"],
            axis=1,
        )

        df_players_data["is_starter"] = df_players_data["position_in_formation"] > 0
        df_players_data.loc[
            df_players_data["is_starter"] & df_players_data["minute_start"].isnull(),
            "minute_start",
        ] = 0
        df_players_data.loc[
            df_players_data["is_starter"] & df_players_data["minute_end"].isnull(), "minute_end"
        ] = game_duration

        df_players_data["minutes_played"] = (
            (df_players_data["minute_end"] - df_players_data["minute_start"]).fillna(0).astype(int)
        )

        players = {}
        for _, player in df_players_data.iterrows():
            if player.minutes_played > 0:
                players[(game_id, player.player_id)] = {
                    # Fields required by the base schema
                    "game_id": game_id,
                    "team_id": player.team_id,
                    "player_id": player.player_id,
                    "player_name": playerid_to_name[player.player_id],
                    "is_starter": player.is_starter,
                    "minutes_played": player.minutes_played,
                    "jersey_number": player.jersey_number,
                    # Fields required by the opta schema
                    "starting_position": self._position_map.get(
                        player.starting_position_id, "Unknown"
                    ),
                }
        return players

    def extract_events(self) -> Dict[Tuple[str, int], Dict[str, Any]]:
        """Return a dictionary with all available events.

        Returns
        -------
        dict
            A mapping between (game ID, event ID) tuples and the information
            available about each event in the data stream.
        """
        match_info = self._get_match_info()
        live_data = self._get_live_data()
        game_id = assertget(match_info, "id")

        events = {}
        for element in assertget(live_data, "event"):
            timestamp_string = assertget(element, "timeStamp")
            timestamp = self._convert_timestamp(timestamp_string)

            qualifiers = {
                int(q["qualifierId"]): q.get("value") for q in element.get("qualifier", [])
            }
            start_x = float(assertget(element, "x"))
            start_y = float(assertget(element, "y"))
            end_x = _get_end_x(qualifiers) or start_x
            end_y = _get_end_y(qualifiers) or start_y

            event_id = int(assertget(element, "id"))
            event = dict(
                # Fields required by the base schema
                game_id=game_id,
                event_id=event_id,
                period_id=int(assertget(element, "periodId")),
                team_id=assertget(element, "contestantId"),
                player_id=element.get("playerId"),
                type_id=int(assertget(element, "typeId")),
                # Fields required by the opta schema
                timestamp=timestamp,
                minute=int(assertget(element, "timeMin")),
                second=int(assertget(element, "timeSec")),
                outcome=bool(int(element.get("outcome", 1))),
                start_x=start_x,
                start_y=start_y,
                end_x=end_x,
                end_y=end_y,
                qualifiers=qualifiers,
                # Optional fields
                assist=bool(int(element.get("assist", 0))),
                keypass=bool(int(element.get("keyPass", 0))),
            )
            events[(game_id, event_id)] = event
        return events

    def extract_substitutions(self) -> Dict[int, Dict[str, Any]]:
        """Return a dictionary with all substitution events.

        Returns
        -------
        dict
            A mapping between player IDs and the information available about
            each substitution in the data stream.
        """
        live_data = self._get_live_data()
        events = assertget(live_data, "event")

        subs = {}
        for e in events:
            event_type = assertget(e, "typeId")
            if event_type in (18, 19):
                sub_id = assertget(e, "playerId")
                substitution_data = {
                    "player_id": assertget(e, "playerId"),
                    "team_id": assertget(e, "contestantId"),
                }
                if event_type == 18:
                    substitution_data["minute_end"] = assertget(e, "timeMin")
                else:
                    substitution_data["minute_start"] = assertget(e, "timeMin")
                subs[sub_id] = substitution_data
        return subs

    def _extract_duration(self) -> int:
        live_data = self._get_live_data()
        events = assertget(live_data, "event")

        game_duration = 90

        for event in events:
            event_type = assertget(event, "typeId")
            if event_type == 30:
                # todo: add 1st half time
                qualifiers = assertget(event, "qualifier")
                for q in qualifiers:
                    qualifier = assertget(q, "qualifierId")
                    if qualifier == 209:
                        new_duration = assertget(event, "timeMin")
                        if new_duration > game_duration:
                            game_duration = new_duration

        return game_duration

    @staticmethod
    def _extract_team_id(teams: List[Dict[str, str]], side: str) -> Optional[str]:
        for team in teams:
            team_side = assertget(team, "position")
            if team_side == side:
                team_id = assertget(team, "id")
                return team_id
        raise MissingDataError

    @staticmethod
    def _convert_timestamp(timestamp_string: str) -> datetime:
        try:
            return datetime.strptime(timestamp_string, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            return datetime.strptime(timestamp_string, "%Y-%m-%dT%H:%M:%SZ")



"""Implements serializers for Wyscout data."""
import glob
import os
import re
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, cast
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen, urlretrieve
from zipfile import ZipFile, is_zipfile

import pandas as pd  # type: ignore
from pandera.typing import DataFrame

from ..base import (
    EventDataLoader,
    JSONType,
    MissingDataError,
    ParseError,
    _auth_remoteloadjson,
    _expand_minute,
    _has_auth,
    _localloadjson,
    _remoteloadjson,
)
from .schema import (
    WyscoutCompetitionSchema,
    WyscoutEventSchema,
    WyscoutGameSchema,
    WyscoutPlayerSchema,
    WyscoutTeamSchema,
)


class PublicWyscoutLoader(EventDataLoader):
    """
    Load the public Wyscout dataset.

    This dataset is a public release of event stream data, collected by Wyscout
    (https://wyscout.com/) containing all matches of the 2017/18 season of the
    top-5 European leagues (La Liga, Serie A, Bundesliga, Premier League, Ligue
    1), the FIFA World Cup 2018, and UEFA Euro Cup 2016. For a detailed
    description, see Pappalardo et al. [1]_.

    Parameters
    ----------
    root : str
        Path where a local copy of the dataset is stored or where the
        downloaded dataset should be stored.
    download : bool
        Whether to force a redownload of the data.

    References
    ----------
    .. [1] Pappalardo, L., Cintia, P., Rossi, A. et al. A public data set of
        spatio-temporal match events in soccer competitions. Sci Data 6, 236
        (2019). https://doi.org/10.1038/s41597-019-0247-7
    """

    def __init__(self, root: Optional[str] = None, download: bool = False) -> None:
        if root is None:
            self.root = os.path.join(os.getcwd(), "wyscout_data")
            os.makedirs(self.root, exist_ok=True)
        else:
            self.root = root

        self.get = _localloadjson

        if download or len(os.listdir(self.root)) == 0:
            self._download_repo()

        self._index = pd.DataFrame(
            [
                {
                    "competition_id": 524,
                    "season_id": 181248,
                    "season_name": "2017/2018",
                    "db_matches": "matches_Italy.json",
                    "db_events": "events_Italy.json",
                },
                {
                    "competition_id": 364,
                    "season_id": 181150,
                    "season_name": "2017/2018",
                    "db_matches": "matches_England.json",
                    "db_events": "events_England.json",
                },
                {
                    "competition_id": 795,
                    "season_id": 181144,
                    "season_name": "2017/2018",
                    "db_matches": "matches_Spain.json",
                    "db_events": "events_Spain.json",
                },
                {
                    "competition_id": 412,
                    "season_id": 181189,
                    "season_name": "2017/2018",
                    "db_matches": "matches_France.json",
                    "db_events": "events_France.json",
                },
                {
                    "competition_id": 426,
                    "season_id": 181137,
                    "season_name": "2017/2018",
                    "db_matches": "matches_Germany.json",
                    "db_events": "events_Germany.json",
                },
                {
                    "competition_id": 102,
                    "season_id": 9291,
                    "season_name": "2016",
                    "db_matches": "matches_European_Championship.json",
                    "db_events": "events_European_Championship.json",
                },
                {
                    "competition_id": 28,
                    "season_id": 10078,
                    "season_name": "2018",
                    "db_matches": "matches_World_Cup.json",
                    "db_events": "events_World_Cup.json",
                },
            ]
        ).set_index(["competition_id", "season_id"])
        self._match_index = self._create_match_index().set_index("match_id")
        self._cache: Optional[Dict[str, Any]] = None

    def _download_repo(self) -> None:
        dataset_urls = dict(
            competitions="https://ndownloader.figshare.com/files/15073685",
            teams="https://ndownloader.figshare.com/files/15073697",
            players="https://ndownloader.figshare.com/files/15073721",
            matches="https://ndownloader.figshare.com/files/14464622",
            events="https://ndownloader.figshare.com/files/14464685",
        )
        # download and unzip Wyscout open data
        for url in dataset_urls.values():
            url_obj = urlopen(url).geturl()
            path = Path(urlparse(url_obj).path)
            file_name = os.path.join(self.root, path.name)
            file_local, _ = urlretrieve(url_obj, file_name)
            if is_zipfile(file_local):
                with ZipFile(file_local) as zip_file:
                    zip_file.extractall(self.root)

    def _create_match_index(self) -> pd.DataFrame:
        df_matches = pd.concat(
            [pd.DataFrame(self.get(path)) for path in glob.iglob(f"{self.root}/matches_*.json")]
        )
        df_matches.rename(
            columns={
                "wyId": "match_id",
                "competitionId": "competition_id",
                "seasonId": "season_id",
            },
            inplace=True,
        )
        return pd.merge(
            df_matches[["match_id", "competition_id", "season_id"]],
            self._index,
            on=["competition_id", "season_id"],
            how="left",
        )

    def competitions(self) -> DataFrame[WyscoutCompetitionSchema]:
        """Return a dataframe with all available competitions and seasons.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available competitions and seasons. See
            :class:`~socceraction.spadl.wyscout.WyscoutCompetitionSchema` for the schema.
        """
        path = os.path.join(self.root, "competitions.json")
        df_competitions = pd.DataFrame(self.get(path))
        df_competitions.rename(
            columns={"wyId": "competition_id", "name": "competition_name"}, inplace=True
        )
        df_competitions["country_name"] = df_competitions.apply(
            lambda x: x.area["name"] if x.area["name"] != "" else "International", axis=1
        )
        df_competitions["competition_gender"] = "male"
        df_competitions = pd.merge(
            df_competitions,
            self._index.reset_index()[["competition_id", "season_id", "season_name"]],
            on="competition_id",
            how="left",
        )
        return cast(
            DataFrame[WyscoutCompetitionSchema],
            df_competitions.reset_index()[
                [
                    "competition_id",
                    "season_id",
                    "country_name",
                    "competition_name",
                    "competition_gender",
                    "season_name",
                ]
            ],
        )

    def games(self, competition_id: int, season_id: int) -> DataFrame[WyscoutGameSchema]:
        """Return a dataframe with all available games in a season.

        Parameters
        ----------
        competition_id : int
            The ID of the competition.
        season_id : int
            The ID of the season.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available games. See
            :class:`~socceraction.spadl.wyscout.WyscoutGameSchema` for the schema.
        """
        path = os.path.join(self.root, self._index.at[(competition_id, season_id), "db_matches"])
        df_matches = pd.DataFrame(self.get(path))
        return cast(DataFrame[WyscoutGameSchema], _convert_games(df_matches))

    def _lineups(self, game_id: int) -> List[Dict[str, Any]]:
        competition_id, season_id = self._match_index.loc[game_id, ["competition_id", "season_id"]]
        path = os.path.join(self.root, self._index.at[(competition_id, season_id), "db_matches"])
        df_matches = pd.DataFrame(self.get(path)).set_index("wyId")
        return list(df_matches.at[game_id, "teamsData"].values())

    def teams(self, game_id: int) -> DataFrame[WyscoutTeamSchema]:
        """Return a dataframe with both teams that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing both teams. See
            :class:`~socceraction.spadl.wyscout.WyscoutTeamSchema` for the schema.
        """
        path = os.path.join(self.root, "teams.json")
        df_teams = pd.DataFrame(self.get(path)).set_index("wyId")
        df_teams_match_id = pd.DataFrame(self._lineups(game_id))["teamId"]
        df_teams_match = df_teams.loc[df_teams_match_id].reset_index()
        return cast(DataFrame[WyscoutTeamSchema], _convert_teams(df_teams_match))

    def players(self, game_id: int) -> DataFrame[WyscoutPlayerSchema]:
        """Return a dataframe with all players that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all players. See
            :class:`~socceraction.spadl.wyscout.WyscoutPlayerSchema` for the schema.
        """
        path = os.path.join(self.root, "players.json")
        df_players = pd.DataFrame(self.get(path)).set_index("wyId")
        lineups = self._lineups(game_id)
        players_match = []
        for team in lineups:
            playerlist = team["formation"]["lineup"]
            if team["formation"]["substitutions"] != "null":
                for p in team["formation"]["substitutions"]:
                    try:
                        playerlist.append(
                            next(
                                item
                                for item in team["formation"]["bench"]
                                if item["playerId"] == p["playerIn"]
                            )
                        )
                    except StopIteration:
                        warnings.warn(
                            f'A player with ID={p["playerIn"]} was substituted '
                            f'in the {p["minute"]}th minute of game {game_id}, but '
                            "could not be found on the bench."
                        )
            df = pd.DataFrame(playerlist)
            df["side"] = team["side"]
            df["team_id"] = team["teamId"]
            players_match.append(df)
        df_players_match = (
            pd.concat(players_match)
            .rename(columns={"playerId": "wyId"})
            .set_index("wyId")
            .join(df_players, how="left")
        )
        df_players_match.reset_index(inplace=True)
        for c in ["shortName", "lastName", "firstName"]:
            df_players_match[c] = df_players_match[c].apply(
                lambda x: x.encode().decode("unicode-escape")
            )
        df_players_match = _convert_players(df_players_match)

        # get minutes played
        competition_id, season_id = self._match_index.loc[game_id, ["competition_id", "season_id"]]
        path = os.path.join(self.root, self._index.at[(competition_id, season_id), "db_events"])
        if self._cache is not None and self._cache["path"] == path:
            df_events = self._cache["events"]
        else:
            df_events = pd.DataFrame(self.get(path)).set_index("matchId")
            # avoid that this large json file has to be parsed again for
            # each game when loading a batch of games from the same season
            self._cache = {"path": path, "events": df_events}
        match_events = df_events.loc[game_id].reset_index().to_dict("records")
        mp = _get_minutes_played(lineups, match_events)
        df_players_match = pd.merge(df_players_match, mp, on="player_id", how="right")
        df_players_match["minutes_played"] = df_players_match.minutes_played.fillna(0)
        df_players_match["game_id"] = game_id
        return cast(DataFrame[WyscoutPlayerSchema], df_players_match)

    def events(self, game_id: int) -> DataFrame[WyscoutEventSchema]:
        """Return a dataframe with the event stream of a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the event stream. See
            :class:`~socceraction.spadl.wyscout.WyscoutEventSchema` for the schema.
        """
        competition_id, season_id = self._match_index.loc[game_id, ["competition_id", "season_id"]]
        path = os.path.join(self.root, self._index.at[(competition_id, season_id), "db_events"])
        if self._cache is not None and self._cache["path"] == path:
            df_events = self._cache["events"]
        else:
            df_events = pd.DataFrame(self.get(path)).set_index("matchId")
            # avoid that this large json file has to be parsed again for
            # each game when loading a batch of games from the same season
            self._cache = {"path": path, "events": df_events}
        return cast(
            DataFrame[WyscoutEventSchema], _convert_events(df_events.loc[game_id].reset_index())
        )


class WyscoutLoader(EventDataLoader):
    """Load event data either from a remote location or from a local folder.

    Parameters
    ----------
    root : str
        Root-path of the data.
    getter : str or callable, default: "remote"
        "remote", "local" or a function that returns loads JSON data from a path.
    feeds : dict(str, str)
        Glob pattern for each feed that should be parsed. The default feeds for
        a "remote" getter are::

            {
                'competitions': 'competitions',
                'seasons': 'competitions/{season_id}/seasons',
                'games': 'seasons/{season_id}/matches',
                'events': 'matches/{game_id}/events?fetch=teams,players,match,substitutions'
            }

        The default feeds for a "local" getter are::

            {
                'competitions': 'competitions.json',
                'seasons': 'seasons_{competition_id}.json',
                'games': 'matches_{season_id}.json',
                'events': 'matches/events_{game_id}.json',
            }

    creds: dict, optional
        Login credentials in the format {"user": "", "passwd": ""}. Only used
        when getter is "remote".
    """

    _wyscout_api: str = "https://apirest.wyscout.com/v2/"

    def __init__(
        self,
        root: str = _wyscout_api,
        getter: Union[str, Callable[[str], JSONType]] = "remote",
        feeds: Optional[Dict[str, str]] = None,
        creds: Optional[Dict[str, str]] = None,
    ) -> None:
        self.root = root

        # Init credentials
        if creds is None:
            creds = {
                "user": os.environ.get("WY_USERNAME", ""),
                "passwd": os.environ.get("WY_PASSWORD", ""),
            }

        # Init getter
        if getter == "remote":
            self.get = _remoteloadjson
            if _has_auth(creds):
                _auth_remoteloadjson(creds["user"], creds["passwd"])
        elif getter == "local":
            self.get = _localloadjson
        else:
            self.get = getter  # type: ignore

        # Set up feeds
        if feeds is not None:
            self.feeds = feeds
        elif getter == "remote":
            self.feeds = {
                "seasons": "competitions/{competition_id}/seasons?fetch=competition",
                "games": "seasons/{season_id}/matches",
                "events": "matches/{game_id}/events?fetch=teams,players,match,coaches,referees,formations,substitutions",  # noqa: B950
            }
        elif getter == "local":
            self.feeds = {
                "competitions": "competitions.json",
                "seasons": "seasons_{competition_id}.json",
                "games": "matches_{season_id}.json",
                "events": "matches/events_{game_id}.json",
            }
        else:
            raise ValueError("No feeds specified.")

    def _get_file_or_url(
        self,
        feed: str,
        competition_id: Optional[int] = None,
        season_id: Optional[int] = None,
        game_id: Optional[int] = None,
    ) -> List[str]:
        competition_id_glob = "*" if competition_id is None else competition_id
        season_id_glob = "*" if season_id is None else season_id
        game_id_glob = "*" if game_id is None else game_id
        glob_pattern = self.feeds[feed].format(
            competition_id=competition_id_glob, season_id=season_id_glob, game_id=game_id_glob
        )
        if "*" in glob_pattern:
            files = glob.glob(os.path.join(self.root, glob_pattern))
            if len(files) == 0:
                raise MissingDataError
            return files
        return [glob_pattern]

    def competitions(
        self, competition_id: Optional[int] = None
    ) -> DataFrame[WyscoutCompetitionSchema]:
        """Return a dataframe with all available competitions and seasons.

        Parameters
        ----------
        competition_id : int, optional
            The ID of the competition.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available competitions and seasons. See
            :class:`~socceraction.spadl.wyscout.WyscoutCompetitionSchema` for the schema.
        """
        # Get all competitions
        if "competitions" in self.feeds:
            competitions_url = self._get_file_or_url("competitions")[0]
            path = os.path.join(self.root, competitions_url)
            obj = self.get(path)
            if not isinstance(obj, dict) or "competitions" not in obj:
                raise ParseError(f"{path} should contain a list of competitions")
            seasons_urls = [
                self._get_file_or_url("seasons", competition_id=c["wyId"])[0]
                for c in obj["competitions"]
            ]
        else:
            seasons_urls = self._get_file_or_url("seasons", competition_id=competition_id)
        # Get seasons in each competition
        competitions = []
        seasons = []
        for seasons_url in seasons_urls:
            try:
                path = os.path.join(self.root, seasons_url)
                obj = self.get(path)
                if not isinstance(obj, dict) or "competition" not in obj or "seasons" not in obj:
                    raise ParseError(
                        f"{path} should contain a list of competition and list of seasons"
                    )
                competitions.append(obj["competition"])
                seasons.extend([s["season"] for s in obj["seasons"]])
            except FileNotFoundError:
                warnings.warn(f"File not found: {seasons_url}")
        df_competitions = _convert_competitions(pd.DataFrame(competitions))
        df_seasons = _convert_seasons(pd.DataFrame(seasons))
        # Merge into a single dataframe
        return cast(
            DataFrame[WyscoutCompetitionSchema],
            pd.merge(df_competitions, df_seasons, on="competition_id"),
        )

    def games(self, competition_id: int, season_id: int) -> DataFrame[WyscoutGameSchema]:
        """Return a dataframe with all available games in a season.

        Parameters
        ----------
        competition_id : int
            The ID of the competition.
        season_id : int
            The ID of the season.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available games. See
            :class:`~socceraction.spadl.wyscout.WyscoutGameSchema` for the schema.
        """
        # Get all games
        if "games" in self.feeds:
            games_url = self._get_file_or_url(
                "games", competition_id=competition_id, season_id=season_id
            )[0]
            path = os.path.join(self.root, games_url)
            obj = self.get(path)
            if not isinstance(obj, dict) or "matches" not in obj:
                raise ParseError(f"{path} should contain a list of matches")
            gamedetails_urls = [
                self._get_file_or_url(
                    "events",
                    competition_id=competition_id,
                    season_id=season_id,
                    game_id=g["matchId"],
                )[0]
                for g in obj["matches"]
            ]
        else:
            gamedetails_urls = self._get_file_or_url(
                "events", competition_id=competition_id, season_id=season_id
            )
        games = []
        for gamedetails_url in gamedetails_urls:
            try:
                path = os.path.join(self.root, gamedetails_url)
                obj = self.get(path)
                if not isinstance(obj, dict) or "match" not in obj:
                    raise ParseError(f"{path} should contain a match")
                games.append(obj["match"])
            except FileNotFoundError:
                warnings.warn(f"File not found: {gamedetails_url}")
            except HTTPError:
                warnings.warn(f"Resource not found: {gamedetails_url}")
        df_games = _convert_games(pd.DataFrame(games))
        return cast(DataFrame[WyscoutGameSchema], df_games)

    def teams(self, game_id: int) -> DataFrame[WyscoutTeamSchema]:
        """Return a dataframe with both teams that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing both teams. See
            :class:`~socceraction.spadl.wyscout.WyscoutTeamSchema` for the schema.
        """
        events_url = self._get_file_or_url("events", game_id=game_id)[0]
        path = os.path.join(self.root, events_url)
        obj = self.get(path)
        if not isinstance(obj, dict) or "teams" not in obj:
            raise ParseError(f"{path} should contain a list of matches")
        teams = [t["team"] for t in obj["teams"].values() if t.get("team")]
        df_teams = _convert_teams(pd.DataFrame(teams))
        return cast(DataFrame[WyscoutTeamSchema], df_teams)

    def players(self, game_id: int) -> DataFrame[WyscoutPlayerSchema]:
        """Return a dataframe with all players that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all players. See
            :class:`~socceraction.spadl.wyscout.WyscoutPlayerSchema` for the schema.
        """
        events_url = self._get_file_or_url("events", game_id=game_id)[0]
        path = os.path.join(self.root, events_url)
        obj = self.get(path)
        if not isinstance(obj, dict) or "players" not in obj:
            raise ParseError(f"{path} should contain a list of players")
        players = [
            player["player"]
            for team in obj["players"].values()
            for player in team
            if player.get("player")
        ]
        df_players = _convert_players(pd.DataFrame(players).drop_duplicates("wyId"))
        df_players = pd.merge(
            df_players,
            _get_minutes_played(obj["match"]["teamsData"], obj["events"]),
            on="player_id",
            how="right",
        )
        df_players["minutes_played"] = df_players.minutes_played.fillna(0)
        df_players["game_id"] = game_id
        return cast(DataFrame[WyscoutPlayerSchema], df_players)

    def events(self, game_id: int) -> DataFrame[WyscoutEventSchema]:
        """Return a dataframe with the event stream of a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the event stream. See
            :class:`~socceraction.spadl.wyscout.WyscoutEventSchema` for the schema.
        """
        events_url = self._get_file_or_url("events", game_id=game_id)[0]
        path = os.path.join(self.root, events_url)
        obj = self.get(path)
        if not isinstance(obj, dict) or "events" not in obj:
            raise ParseError(f"{path} should contain a list of events")
        df_events = _convert_events(pd.DataFrame(obj["events"]))
        return cast(DataFrame[WyscoutEventSchema], df_events)


def _convert_competitions(competitions: pd.DataFrame) -> pd.DataFrame:
    competitionsmapping = {
        "wyId": "competition_id",
        "name": "competition_name",
        "gender": "competition_gender",
    }
    cols = ["competition_id", "competition_name", "country_name", "competition_gender"]
    competitions["country_name"] = competitions.apply(
        lambda x: x.area["name"] if x.area["name"] != "" else "International", axis=1
    )
    competitions = competitions.rename(columns=competitionsmapping)[cols]
    return competitions


def _convert_seasons(seasons: pd.DataFrame) -> pd.DataFrame:
    seasonsmapping = {
        "wyId": "season_id",
        "name": "season_name",
        "competitionId": "competition_id",
    }
    cols = ["season_id", "season_name", "competition_id"]
    seasons = seasons.rename(columns=seasonsmapping)[cols]
    return seasons


def _convert_games(matches: pd.DataFrame) -> pd.DataFrame:
    gamesmapping = {
        "wyId": "game_id",
        "dateutc": "game_date",
        "competitionId": "competition_id",
        "seasonId": "season_id",
        "gameweek": "game_day",
    }
    cols = ["game_id", "competition_id", "season_id", "game_date", "game_day"]
    games = matches.rename(columns=gamesmapping)[cols]
    games["game_date"] = pd.to_datetime(games["game_date"])
    games["home_team_id"] = matches.teamsData.apply(lambda x: _get_team_id(x, "home"))
    games["away_team_id"] = matches.teamsData.apply(lambda x: _get_team_id(x, "away"))
    return games


def _get_team_id(teamsData: Dict[int, Any], side: str) -> int:
    for team_id, data in teamsData.items():
        if data["side"] == side:
            return int(team_id)
    raise ValueError()


def _convert_players(players: pd.DataFrame) -> pd.DataFrame:
    playermapping = {
        "wyId": "player_id",
        "shortName": "nickname",
        "firstName": "firstname",
        "lastName": "lastname",
        "birthDate": "birth_date",
    }
    cols = ["player_id", "nickname", "firstname", "lastname", "birth_date"]
    df_players = players.rename(columns=playermapping)[cols]
    df_players["player_name"] = df_players[["firstname", "lastname"]].agg(" ".join, axis=1)
    df_players["birth_date"] = pd.to_datetime(df_players["birth_date"])
    return df_players


def _convert_teams(teams: pd.DataFrame) -> pd.DataFrame:
    teammapping = {
        "wyId": "team_id",
        "name": "team_name_short",
        "officialName": "team_name",
    }
    cols = ["team_id", "team_name_short", "team_name"]
    return teams.rename(columns=teammapping)[cols]


def _convert_events(raw_events: pd.DataFrame) -> pd.DataFrame:
    eventmapping = {
        "id": "event_id",
        "match_id": "game_id",
        "event_name": "type_name",
        "sub_event_name": "subtype_name",
    }
    cols = [
        "event_id",
        "game_id",
        "period_id",
        "milliseconds",
        "team_id",
        "player_id",
        "type_id",
        "type_name",
        "subtype_id",
        "subtype_name",
        "positions",
        "tags",
    ]
    events = raw_events.copy()
    # Camel case to snake case column names
    pattern = re.compile(r"(?<!^)(?=[A-Z])")
    events.columns = [pattern.sub("_", c).lower() for c in events.columns]
    #
    events["type_id"] = (
        pd.to_numeric(
            events["event_id"] if "event_id" in events.columns else None, errors="coerce"
        )
        .fillna(0)
        .astype(int)
    )
    del events["event_id"]
    events["subtype_id"] = (
        pd.to_numeric(
            events["sub_event_id"] if "sub_event_id" in events.columns else None, errors="coerce"
        )
        .fillna(0)
        .astype(int)
    )
    del events["sub_event_id"]
    events["period_id"] = events.match_period.apply(lambda x: wyscout_periods[x])
    events["milliseconds"] = events.event_sec * 1000
    return events.rename(columns=eventmapping)[cols]


def _get_minutes_played(
    teamsData: List[Dict[str, Any]], events: List[Dict[str, Any]]
) -> pd.DataFrame:
    # get duration of each period
    periods_ts = {i: [0] for i in range(6)}
    for e in events:
        period_id = wyscout_periods[e["matchPeriod"]]
        periods_ts[period_id].append(e["eventSec"])
    periods_duration = [
        round(max(periods_ts[i]) / 60) for i in range(5) if max(periods_ts[i]) != 0
    ]
    # get duration of entire match
    duration = sum(periods_duration)

    # get stats for each player
    playergames: Dict[int, Dict[str, Any]] = {}
    if isinstance(teamsData, dict):
        teamsData = list(teamsData.values())
    for teamData in teamsData:
        formation = teamData.get("formation", {})
        substitutions = formation.get("substitutions", [])
        red_cards = {
            player["playerId"]: _expand_minute(int(player["redCards"]), periods_duration)
            for key in ["bench", "lineup"]
            for player in formation.get(key, [])
            if player["redCards"] != "0"
        }
        pg = {
            player["playerId"]: {
                "team_id": teamData["teamId"],
                "player_id": player["playerId"],
                "jersey_number": player.get("shirtNumber", 0),
                "minutes_played": red_cards.get(player["playerId"], duration),
                "is_starter": True,
            }
            for player in formation.get("lineup", [])
        }

        # correct minutes played for substituted players
        if substitutions != "null":
            for substitution in substitutions:
                expanded_minute_sub = _expand_minute(substitution["minute"], periods_duration)
                substitute = {
                    "team_id": teamData["teamId"],
                    "player_id": substitution["playerIn"],
                    "jersey_number": next(
                        (
                            p.get("shirtNumber", 0)
                            for p in formation.get("bench", [])
                            if p["playerId"] == substitution["playerIn"]
                        ),
                        0,
                    ),
                    "minutes_played": duration - expanded_minute_sub,
                    "is_starter": False,
                }
                if substitution["playerIn"] in red_cards:
                    substitute["minutes_played"] = (
                        red_cards[substitution["playerIn"]] - expanded_minute_sub
                    )
                pg[substitution["playerIn"]] = substitute
                pg[substitution["playerOut"]]["minutes_played"] = expanded_minute_sub

        playergames = {**playergames, **pg}
    return pd.DataFrame(playergames.values())


wyscout_periods = {"1H": 1, "2H": 2, "E1": 3, "E2": 4, "P": 5}



"""Implementation of the Atomic-SPADL language."""

__all__ = [
    'convert_to_atomic',
    'AtomicSPADLSchema',
    'bodyparts_df',
    'actiontypes_df',
    'add_names',
    'play_left_to_right',
]

from .base import convert_to_atomic
from .config import actiontypes_df, bodyparts_df
from .schema import AtomicSPADLSchema
from .utils import add_names, play_left_to_right



"""Base class and utility functions for all event stream data serializers.

A serializer should extend the 'EventDataLoader' class to (down)load event
stream data.
"""
import base64
import json
import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union
from urllib import request

from pandera.typing import DataFrame

JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class ParseError(Exception):
    """Exception raised when a file is not correctly formatted."""


class MissingDataError(Exception):
    """Exception raised when a field is missing in the input data."""


class NoAuthWarning(UserWarning):
    """Warning raised when no user credentials are provided."""


def _remoteloadjson(path: str) -> JSONType:
    """Load JSON data from a URL.

    Parameters
    ----------
    path : str
        URL of the data source.

    Returns
    -------
    JSONType
        A dictionary with the loaded JSON data.
    """
    return json.loads(request.urlopen(path).read())


def _auth_remoteloadjson(user: str, passwd: str) -> None:
    """Add a Authorization header to all requests.

    Parameters
    ----------
    user : str
        Username.
    passwd : str
        Password.
    """
    auth = base64.b64encode(f"{user}:{passwd}".encode())
    opener = request.build_opener()
    opener.addheaders = [("Authorization", f"Basic {auth.decode()}")]
    request.install_opener(opener)


def _localloadjson(path: str) -> JSONType:
    """Load a dictionary from a JSON's filepath.

    Parameters
    ----------
    path : str
        JSON's filepath.

    Returns
    -------
    JSONType
        A dictionary with the data loaded.
    """
    with open(path, encoding="utf-8") as fh:
        return json.load(fh)


def _has_auth(creds: Dict[str, str]) -> bool:
    """Check if user credentials are provided.

    Parameters
    ----------
    creds : dict
        A dictionary with user credentials. It should contain "user" and
        "passwd" keys.

    Returns
    -------
    bool
        True if user credentials are provided, False otherwise.
    """
    if creds.get("user") in [None, ""] or creds.get("passwd") in [None, ""]:
        warnings.warn("Credentials were not supplied. Public data access only.", NoAuthWarning)
        return False
    return True


def _expand_minute(minute: int, periods_duration: List[int]) -> int:
    """Expand a timestamp with injury time of previous periods.

    Parameters
    ----------
    minute : int
        Timestamp in minutes.
    periods_duration : List[int]
        Total duration of each period in minutes.

    Returns
    -------
    int
        Timestamp expanded with injury time.
    """
    expanded_minute = minute
    periods_regular = [45, 45, 15, 15, 0]
    for period in range(len(periods_duration) - 1):
        if minute > sum(periods_regular[: period + 1]):
            expanded_minute += periods_duration[period] - periods_regular[period]
        else:
            break
    return expanded_minute


class EventDataLoader(ABC):
    """Load event data either from a remote location or from a local folder.

    Parameters
    ----------
    root : str
        Root-path of the data.
    getter : str
        "remote" or "local"
    """

    @abstractmethod
    def competitions(self) -> DataFrame[Any]:
        """Return a dataframe with all available competitions and seasons.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available competitions and seasons. See
            :class:`~socceraction.spadl.base.CompetitionSchema` for the schema.
        """

    @abstractmethod
    def games(self, competition_id: int, season_id: int) -> DataFrame[Any]:
        """Return a dataframe with all available games in a season.

        Parameters
        ----------
        competition_id : int
            The ID of the competition.
        season_id : int
            The ID of the season.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available games. See
            :class:`~socceraction.spadl.base.GameSchema` for the schema.
        """

    @abstractmethod
    def teams(self, game_id: int) -> DataFrame[Any]:
        """Return a dataframe with both teams that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing both teams. See
            :class:`~socceraction.spadl.base.TeamSchema` for the schema.
        """

    @abstractmethod
    def players(self, game_id: int) -> DataFrame[Any]:
        """Return a dataframe with all players that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all players. See
            :class:`~socceraction.spadl.base.PlayerSchema` for the schema.
        """

    @abstractmethod
    def events(self, game_id: int) -> DataFrame[Any]:
        """Return a dataframe with the event stream of a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the event stream. See
            :class:`~socceraction.spadl.base.EventSchema` for the schema.
        """



"""Implements serializers for Opta data."""
import copy
import datetime
import glob
import os
import re
import warnings
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Type, Union, cast

import pandas as pd  # type: ignore
from pandera.typing import DataFrame

from socceraction.data.base import EventDataLoader

from .parsers import (
    F1JSONParser,
    F7XMLParser,
    F9JSONParser,
    F24JSONParser,
    F24XMLParser,
    MA1JSONParser,
    MA3JSONParser,
    OptaParser,
    WhoScoredParser,
)
from .schema import (
    OptaCompetitionSchema,
    OptaEventSchema,
    OptaGameSchema,
    OptaPlayerSchema,
    OptaTeamSchema,
)

_jsonparsers = {
    "f1": F1JSONParser,
    "f9": F9JSONParser,
    "f24": F24JSONParser,
    "ma1": MA1JSONParser,
    "ma3": MA3JSONParser,
}

_xmlparsers = {
    "f7": F7XMLParser,
    "f24": F24XMLParser,
}

_statsperformparsers = {
    "ma1": MA1JSONParser,
    "ma3": MA3JSONParser,
}

_whoscoredparsers = {
    "whoscored": WhoScoredParser,
}

_eventtypesdf = pd.DataFrame(
    [
        (1, "pass"),
        (2, "offside pass"),
        (3, "take on"),
        (4, "foul"),
        (5, "out"),
        (6, "corner awarded"),
        (7, "tackle"),
        (8, "interception"),
        (9, "turnover"),
        (10, "save"),
        (11, "claim"),
        (12, "clearance"),
        (13, "miss"),
        (14, "post"),
        (15, "attempt saved"),
        (16, "goal"),
        (17, "card"),
        (18, "player off"),
        (19, "player on"),
        (20, "player retired"),
        (21, "player returns"),
        (22, "player becomes goalkeeper"),
        (23, "goalkeeper becomes player"),
        (24, "condition change"),
        (25, "official change"),
        (26, "unknown26"),
        (27, "start delay"),
        (28, "end delay"),
        (29, "unknown29"),
        (30, "end"),
        (31, "unknown31"),
        (32, "start"),
        (33, "unknown33"),
        (34, "team set up"),
        (35, "player changed position"),
        (36, "player changed jersey number"),
        (37, "collection end"),
        (38, "temp_goal"),
        (39, "temp_attempt"),
        (40, "formation change"),
        (41, "punch"),
        (42, "good skill"),
        (43, "deleted event"),
        (44, "aerial"),
        (45, "challenge"),
        (46, "unknown46"),
        (47, "rescinded card"),
        (48, "unknown46"),
        (49, "ball recovery"),
        (50, "dispossessed"),
        (51, "error"),
        (52, "keeper pick-up"),
        (53, "cross not claimed"),
        (54, "smother"),
        (55, "offside provoked"),
        (56, "shield ball opp"),
        (57, "foul throw in"),
        (58, "penalty faced"),
        (59, "keeper sweeper"),
        (60, "chance missed"),
        (61, "ball touch"),
        (62, "unknown62"),
        (63, "temp_save"),
        (64, "resume"),
        (65, "contentious referee decision"),
        (66, "possession data"),
        (67, "50/50"),
        (68, "referee drop ball"),
        (69, "failed to block"),
        (70, "injury time announcement"),
        (71, "coach setup"),
        (72, "caught offside"),
        (73, "other ball contact"),
        (74, "blocked pass"),
        (75, "delayed start"),
        (76, "early end"),
        (77, "player off pitch"),
        (78, "temp card"),
        (79, "coverage interruption"),
        (80, "drop of ball"),
        (81, "obstacle"),
        (83, "attempted tackle"),
        (84, "deleted after review"),
        (10000, "offside given"),  # Seems specific to WhoScored
    ],
    columns=["type_id", "type_name"],
)


def _deepupdate(target: Dict[Any, Any], src: Dict[Any, Any]) -> None:
    """Deep update target dict with src.

    For each k,v in src: if k doesn't exist in target, it is deep copied from
    src to target. Otherwise, if v is a list, target[k] is extended with
    src[k]. If v is a set, target[k] is updated with v, If v is a dict,
    recursively deep-update it.

    Parameters
    ----------
    target: dict
        The original dictionary which is updated.
    src: dict
        The dictionary with which `target` is updated.

    Examples
    --------
    >>> t = {'name': 'ferry', 'hobbies': ['programming', 'sci-fi']}
    >>> deepupdate(t, {'hobbies': ['gaming']})
    >>> print(t)
    {'name': 'ferry', 'hobbies': ['programming', 'sci-fi', 'gaming']}
    """
    for k, v in src.items():
        if isinstance(v, list):
            if k not in target:
                target[k] = copy.deepcopy(v)
            else:
                target[k].extend(v)
        elif isinstance(v, dict):
            if k not in target:
                target[k] = copy.deepcopy(v)
            else:
                _deepupdate(target[k], v)
        elif isinstance(v, set):
            if k not in target:
                target[k] = v.copy()
            else:
                target[k].update(v.copy())
        else:
            target[k] = copy.copy(v)


def _extract_ids_from_path(path: str, pattern: str) -> Dict[str, Union[str, int]]:
    regex = re.compile(
        ".+?"
        + re.escape(pattern)
        .replace(r"\{competition_id\}", r"(?P<competition_id>[a-zA-Z0-9-_ ]+)")
        .replace(r"\{season_id\}", r"(?P<season_id>[a-zA-Z0-9-_ ]+)")
        .replace(r"\{game_id\}", r"(?P<game_id>[a-zA-Z0-9-_ ]+)")
    )
    m = re.match(regex, path)
    if m is None:
        raise ValueError(f"The filepath {path} does not match the format {pattern}.")
    ids = m.groupdict()
    return {k: int(v) if v.isdigit() else v for k, v in ids.items()}


class OptaLoader(EventDataLoader):
    """Load Opta data feeds from a local folder.

    Parameters
    ----------
    root : str
        Root-path of the data.
    parser : str or dict
        Either 'xml', 'json', 'statsperform', 'whoscored' or a dict with
        a custom parser for each feed. The default xml parser supports F7 and
        F24 feeds; the default json parser supports F1, F9 and F24 feeds, the
        StatsPerform parser supports MA1 and MA3 feeds. Custom parsers can be
        specified as::

            {
                'feed1_name': Feed1Parser
                'feed2_name': Feed2Parser
            }

        where Feed1Parser and Feed2Parser are classes implementing
        :class:`~socceraction.spadl.opta.OptaParser` and 'feed1_name' and
        'feed2_name' are a unique ID for each feed that matches to the keys in
        `feeds`.
    feeds : dict
        Glob pattern describing from which files the data from a specific game
        can be retrieved. For example, if files are named::

            f7-1-2021-17362.xml
            f24-1-2021-17362.xml

        use::

            feeds = {
                'f7': "f7-{competition_id}-{season_id}-{game_id}.xml",
                'f24': "f24-{competition_id}-{season_id}-{game_id}.xml"
            }

    Raises
    ------
    ValueError
        If an invalid parser is provided.
    """

    def __init__(  # noqa: C901
        self,
        root: str,
        parser: Union[str, Mapping[str, Type[OptaParser]]] = "xml",
        feeds: Optional[Dict[str, str]] = None,
    ) -> None:
        self.root = root
        if parser == "json":
            if feeds is None:
                feeds = {
                    "f1": "f1-{competition_id}-{season_id}.json",
                    "f9": "f9-{competition_id}-{season_id}-{game_id}.json",
                    "f24": "f24-{competition_id}-{season_id}-{game_id}.json",
                }
            self.parsers = self._get_parsers_for_feeds(_jsonparsers, feeds)
        elif parser == "xml":
            if feeds is None:
                feeds = {
                    "f7": "f7-{competition_id}-{season_id}-{game_id}.xml",
                    "f24": "f24-{competition_id}-{season_id}-{game_id}.xml",
                }
            self.parsers = self._get_parsers_for_feeds(_xmlparsers, feeds)
        elif parser == "statsperform":
            if feeds is None:
                feeds = {
                    "ma1": "ma1-{competition_id}-{season_id}.json",
                    "ma3": "ma3-{competition_id}-{season_id}-{game_id}.json",
                }
            self.parsers = self._get_parsers_for_feeds(_statsperformparsers, feeds)
        elif parser == "whoscored":
            if feeds is None:
                feeds = {
                    "whoscored": "{competition_id}-{season_id}-{game_id}.json",
                }
            self.parsers = self._get_parsers_for_feeds(_whoscoredparsers, feeds)
        elif isinstance(parser, dict):
            if feeds is None:
                raise ValueError("You must specify a feed for each parser.")
            self.parsers = self._get_parsers_for_feeds(parser, feeds)
        else:
            raise ValueError("Invalid parser provided.")
        self.feeds = {k: str(Path(v)) for k, v in feeds.items()}

    def _get_parsers_for_feeds(
        self, available_parsers: Mapping[str, Type[OptaParser]], feeds: Dict[str, str]
    ) -> Mapping[str, Type[OptaParser]]:
        """Select the appropriate parser for each feed.

        Parameters
        ----------
        available_parsers : dict(str, OptaParser)
            Dictionary with all available parsers.
        feeds : dict(str, str)
            All feeds that should be parsed.

        Returns
        -------
        dict(str, OptaParser)
            A mapping between all feeds that should be parsed and the
            corresponding parser class.

        Warns
        -----
        Raises a warning if there is no parser available for any of the
        provided feeds.
        """
        parsers = {}
        for feed in feeds:
            if feed in available_parsers:
                parsers[feed] = available_parsers[feed]
            else:
                warnings.warn(f"No parser available for {feed} feeds. This feed is ignored.")
        return parsers

    def competitions(self) -> DataFrame[OptaCompetitionSchema]:
        """Return a dataframe with all available competitions and seasons.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available competitions and seasons. See
            :class:`~socceraction.spadl.opta.OptaCompetitionSchema` for the schema.
        """
        data: Dict[int, Dict[str, Any]] = {}
        for feed, feed_pattern in self.feeds.items():
            glob_pattern = feed_pattern.format(competition_id="*", season_id="*", game_id="*")
            feed_files = glob.glob(os.path.join(self.root, glob_pattern))
            for ffp in feed_files:
                ids = _extract_ids_from_path(ffp, feed_pattern)
                parser = self.parsers[feed](ffp, **ids)
                _deepupdate(data, parser.extract_competitions())
        return cast(DataFrame[OptaCompetitionSchema], pd.DataFrame(list(data.values())))

    def games(self, competition_id: int, season_id: int) -> DataFrame[OptaGameSchema]:
        """Return a dataframe with all available games in a season.

        Parameters
        ----------
        competition_id : int
            The ID of the competition.
        season_id : int
            The ID of the season.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available games. See
            :class:`~socceraction.spadl.opta.OptaGameSchema` for the schema.
        """
        data: Dict[int, Dict[str, Any]] = {}
        for feed, feed_pattern in self.feeds.items():
            glob_pattern = feed_pattern.format(
                competition_id=competition_id, season_id=season_id, game_id="*"
            )
            feed_files = glob.glob(os.path.join(self.root, glob_pattern))
            for ffp in feed_files:
                ids = _extract_ids_from_path(ffp, feed_pattern)
                parser = self.parsers[feed](ffp, **ids)
                _deepupdate(data, parser.extract_games())
        return cast(DataFrame[OptaGameSchema], pd.DataFrame(list(data.values())))

    def teams(self, game_id: int) -> DataFrame[OptaTeamSchema]:
        """Return a dataframe with both teams that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing both teams. See
            :class:`~socceraction.spadl.opta.OptaTeamSchema` for the schema.
        """
        data: Dict[int, Dict[str, Any]] = {}
        for feed, feed_pattern in self.feeds.items():
            glob_pattern = feed_pattern.format(competition_id="*", season_id="*", game_id=game_id)
            feed_files = glob.glob(os.path.join(self.root, glob_pattern))
            for ffp in feed_files:
                ids = _extract_ids_from_path(ffp, feed_pattern)
                parser = self.parsers[feed](ffp, **ids)
                _deepupdate(data, parser.extract_teams())
        return cast(DataFrame[OptaTeamSchema], pd.DataFrame(list(data.values())))

    def players(self, game_id: int) -> DataFrame[OptaPlayerSchema]:
        """Return a dataframe with all players that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all players. See
            :class:`~socceraction.spadl.opta.OptaPlayerSchema` for the schema.
        """
        data: Dict[int, Dict[str, Any]] = {}
        for feed, feed_pattern in self.feeds.items():
            glob_pattern = feed_pattern.format(competition_id="*", season_id="*", game_id=game_id)
            feed_files = glob.glob(os.path.join(self.root, glob_pattern))
            for ffp in feed_files:
                ids = _extract_ids_from_path(ffp, feed_pattern)
                parser = self.parsers[feed](ffp, **ids)
                _deepupdate(data, parser.extract_players())
        df_players = pd.DataFrame(list(data.values()))
        df_players["game_id"] = game_id
        return cast(DataFrame[OptaPlayerSchema], df_players)

    def events(self, game_id: int) -> DataFrame[OptaEventSchema]:
        """Return a dataframe with the event stream of a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the event stream. See
            :class:`~socceraction.spadl.opta.OptaEventSchema` for the schema.
        """
        data: Dict[int, Dict[str, Any]] = {}
        for feed, feed_pattern in self.feeds.items():
            glob_pattern = feed_pattern.format(competition_id="*", season_id="*", game_id=game_id)
            feed_files = glob.glob(os.path.join(self.root, glob_pattern))
            for ffp in feed_files:
                ids = _extract_ids_from_path(ffp, feed_pattern)
                parser = self.parsers[feed](ffp, **ids)
                _deepupdate(data, parser.extract_events())
        events = (
            pd.DataFrame(list(data.values()))
            .merge(_eventtypesdf, on="type_id", how="left")
            .sort_values(["game_id", "period_id", "minute", "second", "timestamp"])
            .reset_index(drop=True)
        )

        # sometimes pre-match events has -3, -2 and -1 seconds
        events.loc[events.second < 0, "second"] = 0
        events = events.sort_values(["game_id", "period_id", "minute", "second", "timestamp"])

        # deleted events has wrong datetime which occurs OutOfBoundsDatetime error
        events = events[events.type_id != 43]
        events = events[
            ~(
                (events.timestamp < datetime.datetime(1900, 1, 1))
                | (events.timestamp > datetime.datetime(2100, 1, 1))
            )
        ]

        return cast(DataFrame[OptaEventSchema], events)



"""JSON parser for Opta F24 feeds."""
from datetime import datetime
from typing import Any, Dict, Tuple

from ...base import MissingDataError
from .base import OptaJSONParser, _get_end_x, _get_end_y, assertget


class F24JSONParser(OptaJSONParser):
    """Extract data from a Opta F24 data stream.

    Parameters
    ----------
    path : str
        Path of the data file.
    """

    def _get_doc(self) -> Dict[str, Any]:
        for node in self.root:
            if 'Games' in node['data'].keys():
                return node
        raise MissingDataError

    def extract_games(self) -> Dict[int, Dict[str, Any]]:
        """Return a dictionary with all available games.

        Returns
        -------
        dict
            A mapping between game IDs and the information available about
            each game in the data stream.
        """
        f24 = self._get_doc()

        data = assertget(f24, 'data')
        games = assertget(data, 'Games')
        game = assertget(games, 'Game')
        attr = assertget(game, '@attributes')

        game_id = int(assertget(attr, 'id'))
        game_dict = {
            game_id: dict(
                # Fields required by the base schema
                game_id=game_id,
                season_id=int(assertget(attr, 'season_id')),
                competition_id=int(assertget(attr, 'competition_id')),
                game_day=int(assertget(attr, 'matchday')),
                game_date=datetime.strptime(
                    assertget(assertget(attr, "game_date"), "locale"), "%Y-%m-%dT%H:%M:%S.%fZ"
                ).replace(tzinfo=None),
                home_team_id=int(assertget(attr, 'home_team_id')),
                away_team_id=int(assertget(attr, 'away_team_id')),
                # Fields required by the opta schema
                # home_score=?
                # away_score=?
                # duration=?
                # referee=?
                # venue=?,
                # attendance=?
                # Optional fields
                # home_manager=?
                # away_manager=?
            )
        }
        return game_dict

    def extract_events(self) -> Dict[Tuple[int, int], Dict[str, Any]]:
        """Return a dictionary with all available events.

        Returns
        -------
        dict
            A mapping between (game ID, event ID) tuples and the information
            available about each event in the data stream.
        """
        f24 = self._get_doc()

        data = assertget(f24, 'data')
        games = assertget(data, 'Games')
        game = assertget(games, 'Game')
        game_attr = assertget(game, '@attributes')
        game_id = int(assertget(game_attr, 'id'))

        events = {}
        for element in assertget(game, 'Event'):
            attr = element['@attributes']
            timestamp = attr['TimeStamp'].get('locale') if attr.get('TimeStamp') else None
            timestamp = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%fZ')
            qualifiers = {
                int(q['@attributes']['qualifier_id']): q['@attributes']['value']
                for q in element.get('Q', [])
            }
            start_x = float(assertget(attr, 'x'))
            start_y = float(assertget(attr, 'y'))
            end_x = _get_end_x(qualifiers) or start_x
            end_y = _get_end_y(qualifiers) or start_y

            event_id = int(assertget(attr, 'id'))
            events[(game_id, event_id)] = dict(
                # Fields required by the base schema
                game_id=game_id,
                event_id=event_id,
                period_id=int(assertget(attr, 'period_id')),
                team_id=int(assertget(attr, 'team_id')),
                player_id=int(assertget(attr, 'player_id')),
                type_id=int(assertget(attr, 'type_id')),
                # type_name=?, # added in the opta loader
                # Fields required by the opta schema
                timestamp=timestamp,
                minute=int(assertget(attr, 'min')),
                second=int(assertget(attr, 'sec')),
                outcome=bool(int(attr.get('outcome', 1))),
                start_x=start_x,
                start_y=start_y,
                end_x=end_x,
                end_y=end_y,
                qualifiers=qualifiers,
                # Optional fields
                assist=bool(int(attr.get('assist', 0))),
                keypass=bool(int(attr.get('keypass', 0))),
            )
        return events



"""Module for loading Opta event data."""

__all__ = [
    'OptaLoader',
    'OptaCompetitionSchema',
    'OptaGameSchema',
    'OptaPlayerSchema',
    'OptaTeamSchema',
    'OptaEventSchema',
]

from .loader import OptaLoader
from .schema import (
    OptaCompetitionSchema,
    OptaEventSchema,
    OptaGameSchema,
    OptaPlayerSchema,
    OptaTeamSchema,
)



"""Configuration of the SPADL language.

Attributes
----------
field_length : float
    The length of a pitch (in meters).
field_width : float
    The width of a pitch (in meters).
bodyparts : list(str)
    The bodyparts used in the SPADL language.
results : list(str)
    The action results used in the SPADL language.
actiontypes : list(str)
    The action types used in the SPADL language.

"""
from typing import List

import pandas as pd  # type: ignore

field_length: float = 105.0  # unit: meters
field_width: float = 68.0  # unit: meters

bodyparts: List[str] = ['foot', 'head', 'other', 'head/other', 'foot_left', 'foot_right']
results: List[str] = [
    'fail',
    'success',
    'offside',
    'owngoal',
    'yellow_card',
    'red_card',
]
actiontypes: List[str] = [
    'pass',
    'cross',
    'throw_in',
    'freekick_crossed',
    'freekick_short',
    'corner_crossed',
    'corner_short',
    'take_on',
    'foul',
    'tackle',
    'interception',
    'shot',
    'shot_penalty',
    'shot_freekick',
    'keeper_save',
    'keeper_claim',
    'keeper_punch',
    'keeper_pick_up',
    'clearance',
    'bad_touch',
    'non_action',
    'dribble',
    'goalkick',
]


def actiontypes_df() -> pd.DataFrame:
    """Return a dataframe with the type id and type name of each SPADL action type.

    Returns
    -------
    pd.DataFrame
        The 'type_id' and 'type_name' of each SPADL action type.
    """
    return pd.DataFrame(list(enumerate(actiontypes)), columns=['type_id', 'type_name'])


def results_df() -> pd.DataFrame:
    """Return a dataframe with the result id and result name of each SPADL action type.

    Returns
    -------
    pd.DataFrame
        The 'result_id' and 'result_name' of each SPADL action type.
    """
    return pd.DataFrame(list(enumerate(results)), columns=['result_id', 'result_name'])


def bodyparts_df() -> pd.DataFrame:
    """Return a dataframe with the bodypart id and bodypart name of each SPADL action type.

    Returns
    -------
    pd.DataFrame
        The 'bodypart_id' and 'bodypart_name' of each SPADL action type.
    """
    return pd.DataFrame(list(enumerate(bodyparts)), columns=['bodypart_id', 'bodypart_name'])



"""Implements a converter for regular SPADL actions to atomic actions."""
from typing import cast

import pandas as pd
from pandera.typing import DataFrame

import socceraction.spadl.config as _spadl
from socceraction.spadl.base import _add_dribbles
from socceraction.spadl.schema import SPADLSchema

from . import config as _atomicspadl
from .schema import AtomicSPADLSchema


def convert_to_atomic(actions: DataFrame[SPADLSchema]) -> DataFrame[AtomicSPADLSchema]:
    """Convert regular SPADL actions to atomic actions.

    Parameters
    ----------
    actions : pd.DataFrame
        A SPADL dataframe.

    Returns
    -------
    pd.DataFrame
        The Atomic-SPADL dataframe.
    """
    atomic_actions = cast(pd.DataFrame, actions.copy())
    atomic_actions = _extra_from_passes(atomic_actions)
    atomic_actions = _add_dribbles(atomic_actions)  # for some reason this adds more dribbles
    atomic_actions = _extra_from_shots(atomic_actions)
    atomic_actions = _extra_from_fouls(atomic_actions)
    atomic_actions = _convert_columns(atomic_actions)
    atomic_actions = _simplify(atomic_actions)
    return cast(DataFrame[AtomicSPADLSchema], atomic_actions)


def _extra_from_passes(actions: pd.DataFrame) -> pd.DataFrame:
    next_actions = actions.shift(-1)
    same_team = actions.team_id == next_actions.team_id

    passlike = [
        'pass',
        'cross',
        'throw_in',
        'freekick_short',
        'freekick_crossed',
        'corner_crossed',
        'corner_short',
        'clearance',
        'goalkick',
    ]
    pass_ids = list(_spadl.actiontypes.index(ty) for ty in passlike)

    interceptionlike = [
        'interception',
        'tackle',
        'keeper_punch',
        'keeper_save',
        'keeper_claim',
        'keeper_pick_up',
    ]
    interception_ids = list(_spadl.actiontypes.index(ty) for ty in interceptionlike)

    samegame = actions.game_id == next_actions.game_id
    sameperiod = actions.period_id == next_actions.period_id
    # samephase = next_actions.time_seconds - actions.time_seconds < max_pass_duration
    extra_idx = (
        actions.type_id.isin(pass_ids)
        & samegame
        & sameperiod  # & samephase
        & ~next_actions.type_id.isin(interception_ids)
    )

    prev = actions[extra_idx]
    nex = next_actions[extra_idx]

    extra = pd.DataFrame()
    extra['game_id'] = prev.game_id
    extra['original_event_id'] = prev.original_event_id
    extra['period_id'] = prev.period_id
    extra['action_id'] = prev.action_id + 0.1
    extra['time_seconds'] = (prev.time_seconds + nex.time_seconds) / 2
    extra['start_x'] = prev.end_x
    extra['start_y'] = prev.end_y
    extra['end_x'] = prev.end_x
    extra['end_y'] = prev.end_y
    extra['bodypart_id'] = _atomicspadl.bodyparts.index('foot')
    extra['result_id'] = -1

    offside = prev.result_id == _spadl.results.index('offside')
    out = ((nex.type_id == _atomicspadl.actiontypes.index('goalkick')) & (~same_team)) | (
        nex.type_id == _atomicspadl.actiontypes.index('throw_in')
    )
    ar = _atomicspadl.actiontypes
    extra['type_id'] = -1
    extra['type_id'] = (
        extra.type_id.mask(same_team, ar.index('receival'))
        .mask(~same_team, ar.index('interception'))
        .mask(out, ar.index('out'))
        .mask(offside, ar.index('offside'))
    )
    is_interception = extra['type_id'] == ar.index('interception')
    extra['team_id'] = prev.team_id.mask(is_interception, nex.team_id)
    extra['player_id'] = nex.player_id.mask(out | offside, prev.player_id).astype(
        prev.player_id.dtype
    )

    actions = pd.concat([actions, extra], ignore_index=True, sort=False)
    actions = actions.sort_values(['game_id', 'period_id', 'action_id']).reset_index(drop=True)
    actions['action_id'] = range(len(actions))
    return actions


def _extra_from_shots(actions: pd.DataFrame) -> pd.DataFrame:
    next_actions = actions.shift(-1)

    shotlike = ['shot', 'shot_freekick', 'shot_penalty']
    shot_ids = list(_spadl.actiontypes.index(ty) for ty in shotlike)

    samegame = actions.game_id == next_actions.game_id
    sameperiod = actions.period_id == next_actions.period_id

    shot = actions.type_id.isin(shot_ids)
    goal = shot & (actions.result_id == _spadl.results.index('success'))
    owngoal = actions.result_id == _spadl.results.index('owngoal')
    next_corner_goalkick = next_actions.type_id.isin(
        [
            _atomicspadl.actiontypes.index('corner_crossed'),
            _atomicspadl.actiontypes.index('corner_short'),
            _atomicspadl.actiontypes.index('goalkick'),
        ]
    )
    out = shot & next_corner_goalkick & samegame & sameperiod

    extra_idx = goal | owngoal | out
    prev = actions[extra_idx]
    # nex = next_actions[extra_idx]

    extra = pd.DataFrame()
    extra['game_id'] = prev.game_id
    extra['original_event_id'] = prev.original_event_id
    extra['period_id'] = prev.period_id
    extra['action_id'] = prev.action_id + 0.1
    extra['time_seconds'] = prev.time_seconds  # + nex.time_seconds) / 2
    extra['start_x'] = prev.end_x
    extra['start_y'] = prev.end_y
    extra['end_x'] = prev.end_x
    extra['end_y'] = prev.end_y
    extra['bodypart_id'] = prev.bodypart_id
    extra['result_id'] = -1
    extra['team_id'] = prev.team_id
    extra['player_id'] = prev.player_id

    ar = _atomicspadl.actiontypes
    extra['type_id'] = -1
    extra['type_id'] = (
        extra.type_id.mask(out, ar.index('out'))
        .mask(goal, ar.index('goal'))
        .mask(owngoal, ar.index('owngoal'))
    )
    actions = pd.concat([actions, extra], ignore_index=True, sort=False)
    actions = actions.sort_values(['game_id', 'period_id', 'action_id']).reset_index(drop=True)
    actions['action_id'] = range(len(actions))
    return actions


def _extra_from_fouls(actions: pd.DataFrame) -> pd.DataFrame:
    yellow = actions.result_id == _spadl.results.index('yellow_card')
    red = actions.result_id == _spadl.results.index('red_card')

    prev = actions[yellow | red]
    extra = pd.DataFrame()
    extra['game_id'] = prev.game_id
    extra['original_event_id'] = prev.original_event_id
    extra['period_id'] = prev.period_id
    extra['action_id'] = prev.action_id + 0.1
    extra['time_seconds'] = prev.time_seconds  # + nex.time_seconds) / 2
    extra['start_x'] = prev.end_x
    extra['start_y'] = prev.end_y
    extra['end_x'] = prev.end_x
    extra['end_y'] = prev.end_y
    extra['bodypart_id'] = prev.bodypart_id
    extra['result_id'] = -1
    extra['team_id'] = prev.team_id
    extra['player_id'] = prev.player_id

    ar = _atomicspadl.actiontypes
    extra['type_id'] = -1
    extra['type_id'] = extra.type_id.mask(yellow, ar.index('yellow_card')).mask(
        red, ar.index('red_card')
    )
    actions = pd.concat([actions, extra], ignore_index=True, sort=False)
    actions = actions.sort_values(['game_id', 'period_id', 'action_id']).reset_index(drop=True)
    actions['action_id'] = range(len(actions))
    return actions


def _convert_columns(actions: pd.DataFrame) -> pd.DataFrame:
    actions['x'] = actions.start_x
    actions['y'] = actions.start_y
    actions['dx'] = actions.end_x - actions.start_x
    actions['dy'] = actions.end_y - actions.start_y
    return actions[
        [
            'game_id',
            'original_event_id',
            'action_id',
            'period_id',
            'time_seconds',
            'team_id',
            'player_id',
            'x',
            'y',
            'dx',
            'dy',
            'type_id',
            'bodypart_id',
        ]
    ]


def _simplify(actions: pd.DataFrame) -> pd.DataFrame:
    a = actions
    ar = _atomicspadl.actiontypes

    cornerlike = ['corner_crossed', 'corner_short']
    corner_ids = list(_spadl.actiontypes.index(ty) for ty in cornerlike)

    freekicklike = ['freekick_crossed', 'freekick_short', 'shot_freekick']
    freekick_ids = list(_spadl.actiontypes.index(ty) for ty in freekicklike)

    a['type_id'] = a.type_id.mask(a.type_id.isin(corner_ids), ar.index('corner'))
    a['type_id'] = a.type_id.mask(a.type_id.isin(freekick_ids), ar.index('freekick'))
    return a



"""Base schemas used by all event stream serializers.

Each serializer should create dataframes that contain at least the fields
included in these base schemas. Each serializer can add different additional
fields on top.

"""
import pandas as pd
import pandera as pa
from pandera.typing import DateTime, Object, Series


class CompetitionSchema(pa.SchemaModel):
    """Definition of a dataframe containing a list of competitions and seasons."""

    season_id: Series[Object] = pa.Field()
    """The unique identifier for the season."""
    season_name: Series[str] = pa.Field()
    """The name of the season."""
    competition_id: Series[Object] = pa.Field()
    """The unique identifier for the competition."""
    competition_name: Series[str] = pa.Field()
    """The name of the competition."""

    class Config:  # noqa: D106
        strict = True
        coerce = True


class GameSchema(pa.SchemaModel):
    """Definition of a dataframe containing a list of games."""

    game_id: Series[Object] = pa.Field()
    """The unique identifier for the game."""
    season_id: Series[Object] = pa.Field()
    """The unique identifier for the season."""
    competition_id: Series[Object] = pa.Field()
    """The unique identifier for the competition."""
    game_day: Series[pd.Int64Dtype] = pa.Field(nullable=True)
    """Number corresponding to the weeks or rounds into the competition this game is."""
    game_date: Series[DateTime] = pa.Field()
    """The date when the game was played."""
    home_team_id: Series[Object] = pa.Field()
    """The unique identifier for the home team in this game."""
    away_team_id: Series[Object] = pa.Field()
    """The unique identifier for the away team in this game."""

    class Config:  # noqa: D106
        strict = True
        coerce = True


class TeamSchema(pa.SchemaModel):
    """Definition of a dataframe containing the list of teams of a game."""

    team_id: Series[Object] = pa.Field()
    """The unique identifier for the team."""
    team_name: Series[str] = pa.Field()
    """The name of the team."""

    class Config:  # noqa: D106
        strict = True
        coerce = True


class PlayerSchema(pa.SchemaModel):
    """Definition of a dataframe containing the list of players on the teamsheet of a game."""

    game_id: Series[Object] = pa.Field()
    """The unique identifier for the game."""
    team_id: Series[Object] = pa.Field()
    """The unique identifier for the player's team."""
    player_id: Series[Object] = pa.Field()
    """The unique identifier for the player."""
    player_name: Series[str] = pa.Field()
    """The name of the player."""
    is_starter: Series[bool] = pa.Field()
    """Whether the player is in the starting lineup."""
    minutes_played: Series[int] = pa.Field()
    """The number of minutes the player played in the game."""
    jersey_number: Series[int] = pa.Field()
    """The player's jersey number."""

    class Config:  # noqa: D106
        strict = True
        coerce = True


class EventSchema(pa.SchemaModel):
    """Definition of a dataframe containing event stream data of a game."""

    game_id: Series[Object] = pa.Field()
    """The unique identifier for the game."""
    event_id: Series[Object] = pa.Field()
    """The unique identifier for the event."""
    period_id: Series[int] = pa.Field()
    """The unique identifier for the part of the game in which the event took place."""
    team_id: Series[Object] = pa.Field(nullable=True)
    """The unique identifier for the team this event relates to."""
    player_id: Series[Object] = pa.Field(nullable=True)
    """The unique identifier for the player this event relates to."""
    type_id: Series[int] = pa.Field()
    """The unique identifier for the type of this event."""
    type_name: Series[str] = pa.Field()
    """The name of the type of this event."""

    class Config:  # noqa: D106
        strict = True
        coerce = True



"""Implements serializers for StatsBomb data."""
import os
from typing import Any, Dict, List, Optional, cast

import pandas as pd  # type: ignore
from pandera.typing import DataFrame

try:
    from statsbombpy import sb
except ImportError:
    sb = None

from socceraction.data.base import (
    EventDataLoader,
    ParseError,
    _expand_minute,
    _localloadjson,
)

from .schema import (
    StatsBombCompetitionSchema,
    StatsBombEventSchema,
    StatsBombGameSchema,
    StatsBombPlayerSchema,
    StatsBombTeamSchema,
)


class StatsBombLoader(EventDataLoader):
    """Load Statsbomb data either from a remote location or from a local folder.

    To load remote data, this loader uses the `statsbombpy
    <https://github.com/statsbomb/statsbombpy>`__ package. Data can be retrieved
    from the StatsBomb API and from the `Open Data GitHub repo
    <https://github.com/statsbomb/open-data/>`__.
    API access is for paying customers only. Authentication can be done by
    setting environment variables named ``SB_USERNAME`` and ``SB_PASSWORD`` to
    your login credentials. Alternatively, pass your login credentials using
    the ``creds`` parameter.
    StatsBomb's open data can be accessed without the need of authentication
    but its use is subject to a `user agreement
    <https://github.com/statsbomb/open-data/blob/master/LICENSE.pdf>`__.

    To load local data, point ``root`` to the root folder of the data. This folder
    should use the same directory structure as used in the Open Data GitHub repo.

    Parameters
    ----------
    getter : str
        "remote" or "local"
    root : str, optional
        Root-path of the data. Only used when getter is "local".
    creds: dict, optional
        Login credentials in the format {"user": "", "passwd": ""}. Only used
        when getter is "remote".
    """

    def __init__(
        self,
        getter: str = "remote",
        root: Optional[str] = None,
        creds: Optional[Dict[str, str]] = None,
    ) -> None:
        if getter == "remote":
            if sb is None:
                raise ImportError(
                    """The 'statsbombpy' package is required. Install with 'pip install statsbombpy'."""
                )
            self._creds = creds or sb.DEFAULT_CREDS
            self._local = False
        elif getter == "local":
            if root is None:
                raise ValueError("""The 'root' parameter is required when loading local data.""")
            self._local = True
            self._root = root
        else:
            raise ValueError("Invalid getter specified")

    def competitions(self) -> DataFrame[StatsBombCompetitionSchema]:
        """Return a dataframe with all available competitions and seasons.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available competitions and seasons. See
            :class:`~socceraction.spadl.statsbomb.StatsBombCompetitionSchema` for the schema.
        """
        cols = [
            "season_id",
            "competition_id",
            "competition_name",
            "country_name",
            "competition_gender",
            "season_name",
        ]
        if self._local:
            obj = _localloadjson(str(os.path.join(self._root, "competitions.json")))
        else:
            obj = list(sb.competitions(fmt="dict", creds=self._creds).values())
        if not isinstance(obj, list):
            raise ParseError("The retrieved data should contain a list of competitions")
        if len(obj) == 0:
            return cast(DataFrame[StatsBombCompetitionSchema], pd.DataFrame(columns=cols))
        return cast(DataFrame[StatsBombCompetitionSchema], pd.DataFrame(obj)[cols])

    def games(self, competition_id: int, season_id: int) -> DataFrame[StatsBombGameSchema]:
        """Return a dataframe with all available games in a season.

        Parameters
        ----------
        competition_id : int
            The ID of the competition.
        season_id : int
            The ID of the season.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all available games. See
            :class:`~socceraction.spadl.statsbomb.StatsBombGameSchema` for the schema.
        """
        cols = [
            "game_id",
            "season_id",
            "competition_id",
            "competition_stage",
            "game_day",
            "game_date",
            "home_team_id",
            "away_team_id",
            "home_score",
            "away_score",
            "venue",
            "referee",
        ]
        if self._local:
            obj = _localloadjson(
                str(os.path.join(self._root, f"matches/{competition_id}/{season_id}.json"))
            )
        else:
            obj = list(
                sb.matches(competition_id, season_id, fmt="dict", creds=self._creds).values()
            )
        if not isinstance(obj, list):
            raise ParseError("The retrieved data should contain a list of games")
        if len(obj) == 0:
            return cast(DataFrame[StatsBombGameSchema], pd.DataFrame(columns=cols))
        gamesdf = pd.DataFrame(_flatten(m) for m in obj)
        gamesdf["kick_off"] = gamesdf["kick_off"].fillna("12:00:00.000")
        gamesdf["match_date"] = pd.to_datetime(
            gamesdf[["match_date", "kick_off"]].agg(" ".join, axis=1)
        )
        gamesdf.rename(
            columns={
                "match_id": "game_id",
                "match_date": "game_date",
                "match_week": "game_day",
                "stadium_name": "venue",
                "referee_name": "referee",
                "competition_stage_name": "competition_stage",
            },
            inplace=True,
        )
        if "venue" not in gamesdf:
            gamesdf["venue"] = None
        if "referee" not in gamesdf:
            gamesdf["referee"] = None
        return cast(DataFrame[StatsBombGameSchema], gamesdf[cols])

    def _lineups(self, game_id: int) -> List[Dict[str, Any]]:
        if self._local:
            obj = _localloadjson(str(os.path.join(self._root, f"lineups/{game_id}.json")))
        else:
            obj = list(sb.lineups(game_id, fmt="dict", creds=self._creds).values())
        if not isinstance(obj, list):
            raise ParseError("The retrieved data should contain a list of teams")
        if len(obj) != 2:
            raise ParseError("The retrieved data should contain two teams")
        return obj

    def teams(self, game_id: int) -> DataFrame[StatsBombTeamSchema]:
        """Return a dataframe with both teams that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Raises
        ------
        ParseError  # noqa: DAR402
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing both teams. See
            :class:`~socceraction.spadl.statsbomb.StatsBombTeamSchema` for the schema.
        """
        cols = ["team_id", "team_name"]
        obj = self._lineups(game_id)
        return cast(DataFrame[StatsBombTeamSchema], pd.DataFrame(obj)[cols])

    def players(self, game_id: int) -> DataFrame[StatsBombPlayerSchema]:
        """Return a dataframe with all players that participated in a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.

        Raises
        ------
        ParseError  # noqa: DAR402
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing all players. See
            :class:`~socceraction.spadl.statsbomb.StatsBombPlayerSchema` for the schema.
        """
        cols = [
            "game_id",
            "team_id",
            "player_id",
            "player_name",
            "nickname",
            "jersey_number",
            "is_starter",
            "starting_position_id",
            "starting_position_name",
            "minutes_played",
        ]

        obj = self._lineups(game_id)
        playersdf = pd.DataFrame(_flatten_id(p) for lineup in obj for p in lineup["lineup"])
        playergamesdf = extract_player_games(self.events(game_id))
        playersdf = pd.merge(
            playersdf,
            playergamesdf[
                ["player_id", "team_id", "position_id", "position_name", "minutes_played"]
            ],
            on="player_id",
        )
        playersdf["game_id"] = game_id
        playersdf["position_name"] = playersdf["position_name"].replace(0, "Substitute")
        playersdf["position_id"] = playersdf["position_id"].fillna(0).astype(int)
        playersdf["is_starter"] = playersdf["position_id"] != 0
        playersdf.rename(
            columns={
                "player_nickname": "nickname",
                "country_name": "country",
                "position_id": "starting_position_id",
                "position_name": "starting_position_name",
            },
            inplace=True,
        )
        return cast(DataFrame[StatsBombPlayerSchema], playersdf[cols])

    def events(self, game_id: int, load_360: bool = False) -> DataFrame[StatsBombEventSchema]:
        """Return a dataframe with the event stream of a game.

        Parameters
        ----------
        game_id : int
            The ID of the game.
        load_360 : bool
            Whether to load the 360 data.

        Raises
        ------
        ParseError
            When the raw data does not adhere to the expected format.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the event stream. See
            :class:`~socceraction.spadl.statsbomb.StatsBombEventSchema` for the schema.
        """
        cols = [
            "game_id",
            "event_id",
            "period_id",
            "team_id",
            "player_id",
            "type_id",
            "type_name",
            "index",
            "timestamp",
            "minute",
            "second",
            "possession",
            "possession_team_id",
            "possession_team_name",
            "play_pattern_id",
            "play_pattern_name",
            "team_name",
            "duration",
            "extra",
            "related_events",
            "player_name",
            "position_id",
            "position_name",
            "location",
            "under_pressure",
            "counterpress",
        ]
        # Load the events
        if self._local:
            obj = _localloadjson(str(os.path.join(self._root, f"events/{game_id}.json")))
        else:
            obj = list(sb.events(game_id, fmt="dict", creds=self._creds).values())
        if not isinstance(obj, list):
            raise ParseError("The retrieved data should contain a list of events")
        if len(obj) == 0:
            return cast(DataFrame[StatsBombEventSchema], pd.DataFrame(columns=cols))

        eventsdf = pd.DataFrame(_flatten_id(e) for e in obj)
        eventsdf["match_id"] = game_id
        eventsdf["timestamp"] = pd.to_datetime(eventsdf["timestamp"], format="%H:%M:%S.%f")
        eventsdf["related_events"] = eventsdf["related_events"].apply(
            lambda d: d if isinstance(d, list) else []
        )
        eventsdf["under_pressure"] = eventsdf["under_pressure"].fillna(False).astype(bool)
        eventsdf["counterpress"] = eventsdf["counterpress"].fillna(False).astype(bool)
        eventsdf.rename(
            columns={"id": "event_id", "period": "period_id", "match_id": "game_id"},
            inplace=True,
        )
        if not load_360:
            return cast(DataFrame[StatsBombEventSchema], eventsdf[cols])

        # Load the 360 data
        cols_360 = ["visible_area_360", "freeze_frame_360"]
        if self._local:
            obj = _localloadjson(str(os.path.join(self._root, f"three-sixty/{game_id}.json")))
        else:
            obj = sb.frames(game_id, fmt="dict", creds=self._creds)
        if not isinstance(obj, list):
            raise ParseError("The retrieved data should contain a list of frames")
        if len(obj) == 0:
            eventsdf["visible_area_360"] = None
            eventsdf["freeze_frame_360"] = None
            return cast(DataFrame[StatsBombEventSchema], eventsdf[cols + cols_360])
        framesdf = pd.DataFrame(obj).rename(
            columns={
                "event_uuid": "event_id",
                "visible_area": "visible_area_360",
                "freeze_frame": "freeze_frame_360",
            },
        )[["event_id", "visible_area_360", "freeze_frame_360"]]
        return cast(
            DataFrame[StatsBombEventSchema],
            pd.merge(eventsdf, framesdf, on="event_id", how="left")[cols + cols_360],
        )


def extract_player_games(events: pd.DataFrame) -> pd.DataFrame:
    """Extract player games [player_id, game_id, minutes_played] from statsbomb match events.

    Parameters
    ----------
    events : pd.DataFrame
        DataFrame containing StatsBomb events of a single game.

    Returns
    -------
    player_games : pd.DataFrame
        A DataFrame with the number of minutes played by each player during the game.
    """
    # get duration of each period
    periods = pd.DataFrame(
        [
            {"period_id": 1, "minute": 45},
            {"period_id": 2, "minute": 45},
            {"period_id": 3, "minute": 15},
            {"period_id": 4, "minute": 15},
            # Shoot-outs should not contritbute to minutes played
            # {"period_id": 5, "minute": 0},
        ]
    ).set_index("period_id")
    periods_minutes = (
        events.loc[events.type_name == "Half End", ["period_id", "minute"]]
        .drop_duplicates()
        .set_index("period_id")
        .sort_index()
        .subtract(periods.cumsum().shift(1).fillna(0))
        .minute.dropna()
        .astype(int)
        .tolist()
    )
    # get duration of entire match
    game_minutes = sum(periods_minutes)

    game_id = events.game_id.mode().values[0]
    players = {}
    # Red cards
    red_cards = events[
        events.apply(
            lambda x: any(
                e in x.extra
                and "card" in x.extra[e]
                and x.extra[e]["card"]["name"] in ["Second Yellow", "Red Card"]
                for e in ["foul_committed", "bad_behaviour"]
            ),
            axis=1,
        )
    ]
    # stats for starting XI
    for startxi in events[events.type_name == "Starting XI"].itertuples():
        team_id, team_name = startxi.team_id, startxi.team_name
        for player in startxi.extra["tactics"]["lineup"]:
            player = _flatten_id(player)
            player = {
                **player,
                **{
                    "game_id": game_id,
                    "team_id": team_id,
                    "team_name": team_name,
                    "minutes_played": game_minutes,
                },
            }
            player_red_card = red_cards[red_cards.player_id == player["player_id"]]
            if len(player_red_card) > 0:
                red_card_minute = player_red_card.iloc[0].minute
                player["minutes_played"] = _expand_minute(red_card_minute, periods_minutes)
            players[player["player_id"]] = player
    # stats for substitutions
    for substitution in events[events.type_name == "Substitution"].itertuples():
        exp_sub_minute = _expand_minute(substitution.minute, periods_minutes)
        replacement = {
            "player_id": substitution.extra["substitution"]["replacement"]["id"],
            "player_name": substitution.extra["substitution"]["replacement"]["name"],
            "minutes_played": game_minutes - exp_sub_minute,
            "team_id": substitution.team_id,
            "game_id": game_id,
            "team_name": substitution.team_name,
        }
        player_red_card = red_cards[red_cards.player_id == replacement["player_id"]]
        if len(player_red_card) > 0:
            red_card_minute = player_red_card.iloc[0].minute
            replacement["minutes_played"] = (
                _expand_minute(red_card_minute, periods_minutes) - exp_sub_minute
            )
        players[replacement["player_id"]] = replacement
        players[substitution.player_id]["minutes_played"] = exp_sub_minute
    pg = pd.DataFrame(players.values()).fillna(0)
    for col in pg.columns:
        if "_id" in col:
            pg[col] = pg[col].astype(int)  # pylint: disable=E1136,E1137
    return pg


def _flatten_id(d: Dict[str, Dict[str, Any]]) -> Dict[str, Any]:
    newd = {}
    extra = {}
    for k, v in d.items():
        if isinstance(v, dict):
            if "id" in v and "name" in v:
                newd[k + "_id"] = v["id"]
                newd[k + "_name"] = v["name"]
            else:
                extra[k] = v
        else:
            newd[k] = v
    newd["extra"] = extra
    return newd


def _flatten(d: Dict[str, Dict[str, Any]]) -> Dict[str, Any]:
    newd = {}
    for k, v in d.items():
        if isinstance(v, dict):
            if "id" in v and "name" in v:
                newd[k + "_id"] = v["id"]
                newd[k + "_name"] = v["name"]
                newd[k + "_extra"] = {l: w for (l, w) in v.items() if l in ("id", "name")}
            else:
                newd = {**newd, **_flatten(v)}
        else:
            newd[k] = v
    return newd



"""Module for loading Wyscout event data."""

__all__ = [
    'PublicWyscoutLoader',
    'WyscoutLoader',
    'WyscoutCompetitionSchema',
    'WyscoutGameSchema',
    'WyscoutPlayerSchema',
    'WyscoutTeamSchema',
    'WyscoutEventSchema',
]

from .loader import PublicWyscoutLoader, WyscoutLoader
from .schema import (
    WyscoutCompetitionSchema,
    WyscoutEventSchema,
    WyscoutGameSchema,
    WyscoutPlayerSchema,
    WyscoutTeamSchema,
)



"""Wyscout event stream data to SPADL converter."""
from typing import Any, Dict, List, Optional, Set, cast

import pandas as pd  # type: ignore
from pandera.typing import DataFrame

from . import config as spadlconfig
from .base import (
    _add_dribbles,
    _fix_clearances,
    _fix_direction_of_play,
    min_dribble_length,
)
from .schema import SPADLSchema

###################################
# WARNING: HERE BE DRAGONS
# This code for converting wyscout data was organically grown over a long period of time.
# It works for now, but needs to be cleaned up in the future.
# Enter at your own risk.
###################################


def convert_to_actions(events: pd.DataFrame, home_team_id: int) -> DataFrame[SPADLSchema]:
    """
    Convert Wyscout events to SPADL actions.

    Parameters
    ----------
    events : pd.DataFrame
        DataFrame containing Wyscout events from a single game.
    home_team_id : int
        ID of the home team in the corresponding game.

    Returns
    -------
    actions : pd.DataFrame
        DataFrame with corresponding SPADL actions.

    """
    events = pd.concat([events, get_tagsdf(events)], axis=1)
    events = make_new_positions(events)
    events = fix_wyscout_events(events)
    actions = create_df_actions(events)
    actions = fix_actions(actions)
    actions = _fix_direction_of_play(actions, home_team_id)
    actions = _fix_clearances(actions)
    actions["action_id"] = range(len(actions))
    actions = _add_dribbles(actions)

    return cast(DataFrame[SPADLSchema], actions)


def _get_tag_set(tags: List[Dict[str, Any]]) -> Set[int]:
    return {tag["id"] for tag in tags}


def get_tagsdf(events: pd.DataFrame) -> pd.DataFrame:
    """Represent Wyscout tags as a boolean dataframe.

    Parameters
    ----------
    events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        A dataframe with a column for each tag.
    """
    tags = events.tags.apply(_get_tag_set)
    tagsdf = pd.DataFrame()
    for tag_id, column in wyscout_tags:
        tagsdf[column] = tags.apply(lambda x, tag=tag_id: tag in x)
    return tagsdf


wyscout_tags = [
    (101, "goal"),
    (102, "own_goal"),
    (301, "assist"),
    (302, "key_pass"),
    (1901, "counter_attack"),
    (401, "left_foot"),
    (402, "right_foot"),
    (403, "head/body"),
    (1101, "direct"),
    (1102, "indirect"),
    (2001, "dangerous_ball_lost"),
    (2101, "blocked"),
    (801, "high"),
    (802, "low"),
    (1401, "interception"),
    (1501, "clearance"),
    (201, "opportunity"),
    (1301, "feint"),
    (1302, "missed_ball"),
    (501, "free_space_right"),
    (502, "free_space_left"),
    (503, "take_on_left"),
    (504, "take_on_right"),
    (1601, "sliding_tackle"),
    (601, "anticipated"),
    (602, "anticipation"),
    (1701, "red_card"),
    (1702, "yellow_card"),
    (1703, "second_yellow_card"),
    (1201, "position_goal_low_center"),
    (1202, "position_goal_low_right"),
    (1203, "position_goal_mid_center"),
    (1204, "position_goal_mid_left"),
    (1205, "position_goal_low_left"),
    (1206, "position_goal_mid_right"),
    (1207, "position_goal_high_center"),
    (1208, "position_goal_high_left"),
    (1209, "position_goal_high_right"),
    (1210, "position_out_low_right"),
    (1211, "position_out_mid_left"),
    (1212, "position_out_low_left"),
    (1213, "position_out_mid_right"),
    (1214, "position_out_high_center"),
    (1215, "position_out_high_left"),
    (1216, "position_out_high_right"),
    (1217, "position_post_low_right"),
    (1218, "position_post_mid_left"),
    (1219, "position_post_low_left"),
    (1220, "position_post_mid_right"),
    (1221, "position_post_high_center"),
    (1222, "position_post_high_left"),
    (1223, "position_post_high_right"),
    (901, "through"),
    (1001, "fairplay"),
    (701, "lost"),
    (702, "neutral"),
    (703, "won"),
    (1801, "accurate"),
    (1802, "not_accurate"),
]


def _make_position_vars(event_id: int, positions: List[Dict[str, Optional[float]]]) -> pd.Series:
    if len(positions) == 2:  # if less than 2 then action is removed
        start_x = positions[0]["x"]
        start_y = positions[0]["y"]
        end_x = positions[1]["x"]
        end_y = positions[1]["y"]
    elif len(positions) == 1:
        start_x = positions[0]["x"]
        start_y = positions[0]["y"]
        end_x = start_x
        end_y = start_y
    else:
        start_x = None
        start_y = None
        end_x = None
        end_y = None
    return pd.Series([event_id, start_x, start_y, end_x, end_y])


def make_new_positions(events: pd.DataFrame) -> pd.DataFrame:
    """Extract the start and end coordinates for each action.

    Parameters
    ----------
    events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe with start and end coordinates for each action.
    """
    new_positions = events[["event_id", "positions"]].apply(
        lambda x: _make_position_vars(x[0], x[1]), axis=1
    )
    new_positions.columns = ["event_id", "start_x", "start_y", "end_x", "end_y"]
    events = pd.merge(events, new_positions, left_on="event_id", right_on="event_id")
    events[["start_x", "end_x"]] = events[["start_x", "end_x"]]
    events[["start_y", "end_y"]] = events[["start_y", "end_y"]]
    events = events.drop("positions", axis=1)
    return events


def fix_wyscout_events(df_events: pd.DataFrame) -> pd.DataFrame:
    """Perform some fixes on the Wyscout events such that the spadl action dataframe can be built.

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe with an extra column 'offside'
    """
    df_events = create_shot_coordinates(df_events)
    df_events = convert_duels(df_events)
    df_events = insert_interception_passes(df_events)
    df_events = add_offside_variable(df_events)
    df_events = convert_touches(df_events)
    df_events = convert_simulations(df_events)
    return df_events


def create_shot_coordinates(df_events: pd.DataFrame) -> pd.DataFrame:
    """Create shot coordinates (estimates) from the Wyscout tags.

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe with end coordinates for shots
    """
    goal_center_idx = (
        df_events["position_goal_low_center"]
        | df_events["position_goal_mid_center"]
        | df_events["position_goal_high_center"]
    )
    df_events.loc[goal_center_idx, "end_x"] = 100.0
    df_events.loc[goal_center_idx, "end_y"] = 50.0

    goal_right_idx = (
        df_events["position_goal_low_right"]
        | df_events["position_goal_mid_right"]
        | df_events["position_goal_high_right"]
    )
    df_events.loc[goal_right_idx, "end_x"] = 100.0
    df_events.loc[goal_right_idx, "end_y"] = 55.0

    goal_left_idx = (
        df_events["position_goal_mid_left"]
        | df_events["position_goal_low_left"]
        | df_events["position_goal_high_left"]
    )
    df_events.loc[goal_left_idx, "end_x"] = 100.0
    df_events.loc[goal_left_idx, "end_y"] = 45.0

    out_center_idx = df_events["position_out_high_center"] | df_events["position_post_high_center"]
    df_events.loc[out_center_idx, "end_x"] = 100.0
    df_events.loc[out_center_idx, "end_y"] = 50.0

    out_right_idx = (
        df_events["position_out_low_right"]
        | df_events["position_out_mid_right"]
        | df_events["position_out_high_right"]
    )
    df_events.loc[out_right_idx, "end_x"] = 100.0
    df_events.loc[out_right_idx, "end_y"] = 60.0

    out_left_idx = (
        df_events["position_out_mid_left"]
        | df_events["position_out_low_left"]
        | df_events["position_out_high_left"]
    )
    df_events.loc[out_left_idx, "end_x"] = 100.0
    df_events.loc[out_left_idx, "end_y"] = 40.0

    post_left_idx = (
        df_events["position_post_mid_left"]
        | df_events["position_post_low_left"]
        | df_events["position_post_high_left"]
    )
    df_events.loc[post_left_idx, "end_x"] = 100.0
    df_events.loc[post_left_idx, "end_y"] = 55.38

    post_right_idx = (
        df_events["position_post_low_right"]
        | df_events["position_post_mid_right"]
        | df_events["position_post_high_right"]
    )
    df_events.loc[post_right_idx, "end_x"] = 100.0
    df_events.loc[post_right_idx, "end_y"] = 44.62

    blocked_idx = df_events["blocked"]
    df_events.loc[blocked_idx, "end_x"] = df_events.loc[blocked_idx, "start_x"]
    df_events.loc[blocked_idx, "end_y"] = df_events.loc[blocked_idx, "start_y"]

    return df_events


def convert_duels(df_events: pd.DataFrame) -> pd.DataFrame:
    """Convert duel events.

    This function converts Wyscout duels that end with the ball out of field
    (subtype_id 50) into a pass for the player winning the duel to the location
    of where the ball went out of field. The remaining duels are removed as
    they are not on-the-ball actions.

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe in which the duels are either removed or
        transformed into a pass
    """
    # Shift events dataframe by one and two time steps
    df_events1 = df_events.shift(-1)
    df_events2 = df_events.shift(-2)

    # Define selector for same period id
    selector_same_period = df_events["period_id"] == df_events2["period_id"]

    # Define selector for duels that are followed by an 'out of field' event
    selector_duel_out_of_field = (
        (df_events["type_id"] == 1)
        & (df_events1["type_id"] == 1)
        & (df_events2["subtype_id"] == 50)
        & selector_same_period
    )

    # Define selectors for current time step
    selector0_duel_won = selector_duel_out_of_field & (
        df_events["team_id"] != df_events2["team_id"]
    )
    selector0_duel_won_air = selector0_duel_won & (df_events["subtype_id"] == 10)
    selector0_duel_won_not_air = selector0_duel_won & (df_events["subtype_id"] != 10)

    # Define selectors for next time step
    selector1_duel_won = selector_duel_out_of_field & (
        df_events1["team_id"] != df_events2["team_id"]
    )
    selector1_duel_won_air = selector1_duel_won & (df_events1["subtype_id"] == 10)
    selector1_duel_won_not_air = selector1_duel_won & (df_events1["subtype_id"] != 10)

    # Aggregate selectors
    selector_duel_won = selector0_duel_won | selector1_duel_won
    selector_duel_won_air = selector0_duel_won_air | selector1_duel_won_air
    selector_duel_won_not_air = selector0_duel_won_not_air | selector1_duel_won_not_air

    # Set types and subtypes
    df_events.loc[selector_duel_won, "type_id"] = 8
    df_events.loc[selector_duel_won_air, "subtype_id"] = 82
    df_events.loc[selector_duel_won_not_air, "subtype_id"] = 85

    # set end location equal to ball out of field location
    df_events.loc[selector_duel_won, "accurate"] = False
    df_events.loc[selector_duel_won, "not_accurate"] = True
    df_events.loc[selector_duel_won, "end_x"] = 100 - df_events2.loc[selector_duel_won, "start_x"]
    df_events.loc[selector_duel_won, "end_y"] = 100 - df_events2.loc[selector_duel_won, "start_y"]

    # df_events.loc[selector_duel_won, 'end_x'] = df_events2.loc[selector_duel_won, 'start_x']
    # df_events.loc[selector_duel_won, 'end_y'] = df_events2.loc[selector_duel_won, 'start_y']

    # Define selector for ground attacking duels with take on
    selector_attacking_duel = df_events["subtype_id"] == 11
    selector_take_on = (df_events["take_on_left"]) | (df_events["take_on_right"])
    selector_att_duel_take_on = selector_attacking_duel & selector_take_on

    # Set take ons type to 0
    df_events.loc[selector_att_duel_take_on, "type_id"] = 0

    # Set sliding tackles type to 0
    df_events.loc[df_events["sliding_tackle"], "type_id"] = 0

    # Remove the remaining duels
    df_events = df_events[df_events["type_id"] != 1]

    # Reset the index
    df_events = df_events.reset_index(drop=True)

    return df_events


def insert_interception_passes(df_events: pd.DataFrame) -> pd.DataFrame:
    """Insert interception actions before passes.

    This function converts passes (type_id 8) that are also interceptions
    (tag interception) in the Wyscout event data into two separate events,
    first an interception and then a pass.

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe in which passes that were also denoted as
        interceptions in the Wyscout notation are transformed into two events
    """
    df_events_interceptions = df_events[
        df_events["interception"] & (df_events["type_id"] == 8)
    ].copy()

    if not df_events_interceptions.empty:
        df_events_interceptions.loc[:, [t[1] for t in wyscout_tags]] = False
        df_events_interceptions["interception"] = True
        df_events_interceptions["type_id"] = 0
        df_events_interceptions["subtype_id"] = 0
        df_events_interceptions[["end_x", "end_y"]] = df_events_interceptions[
            ["start_x", "start_y"]
        ]

        df_events = pd.concat([df_events_interceptions, df_events], ignore_index=True)
        df_events = df_events.sort_values(["period_id", "milliseconds"])
        df_events = df_events.reset_index(drop=True)

    return df_events


def add_offside_variable(df_events: pd.DataFrame) -> pd.DataFrame:
    """Attach offside events to the previous action.

    This function removes the offside events in the Wyscout event data and adds
    sets offside to 1 for the previous event (if this was a passing event)

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe with an extra column 'offside'
    """
    # Create a new column for the offside variable
    df_events["offside"] = 0

    # Shift events dataframe by one timestep
    df_events1 = df_events.shift(-1)

    # Select offside passes
    selector_offside = (df_events1["type_id"] == 6) & (df_events["type_id"] == 8)

    # Set variable 'offside' to 1 for all offside passes
    df_events.loc[selector_offside, "offside"] = 1

    # Remove offside events
    df_events = df_events[df_events["type_id"] != 6]

    # Reset index
    df_events = df_events.reset_index(drop=True)

    return df_events


def convert_simulations(df_events: pd.DataFrame) -> pd.DataFrame:
    """Convert simulations to failed take-ons.

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe


    Returns
    -------
        pd.DataFrame
        Wyscout event dataframe in which simulation events are either
        transformed into a failed take-on
    """
    prev_events = df_events.shift(1)

    # Select simulations
    selector_simulation = df_events["subtype_id"] == 25

    # Select actions preceded by a failed take-on
    selector_previous_is_failed_take_on = (prev_events["take_on_left"]) | (
        prev_events["take_on_right"]
    ) & prev_events["not_accurate"]

    # Transform simulations not preceded by a failed take-on to a failed take-on
    df_events.loc[selector_simulation & ~selector_previous_is_failed_take_on, "type_id"] = 0
    df_events.loc[selector_simulation & ~selector_previous_is_failed_take_on, "subtype_id"] = 0
    df_events.loc[selector_simulation & ~selector_previous_is_failed_take_on, "accurate"] = False
    df_events.loc[
        selector_simulation & ~selector_previous_is_failed_take_on, "not_accurate"
    ] = True
    # Set take_on_left or take_on_right to True
    df_events.loc[
        selector_simulation & ~selector_previous_is_failed_take_on, "take_on_left"
    ] = True

    # Remove simulation events which are preceded by a failed take-on
    df_events = df_events[~(selector_simulation & selector_previous_is_failed_take_on)]

    # Reset index
    df_events = df_events.reset_index(drop=True)

    return df_events


def convert_touches(df_events: pd.DataFrame) -> pd.DataFrame:
    """Convert touch events to dribbles or passes.

    This function converts the Wyscout 'touch' event (sub_type_id 72) into either
    a dribble or a pass (accurate or not depending on receiver)

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        Wyscout event dataframe without any touch events
    """
    df_events1 = df_events.shift(-1)

    selector_touch = (df_events["subtype_id"] == 72) & ~df_events["interception"]

    selector_same_player = df_events["player_id"] == df_events1["player_id"]
    selector_same_team = df_events["team_id"] == df_events1["team_id"]

    # selector_touch_same_player = selector_touch & selector_same_player
    selector_touch_same_team = selector_touch & ~selector_same_player & selector_same_team
    selector_touch_other = selector_touch & ~selector_same_player & ~selector_same_team

    same_x = abs(df_events["end_x"] - df_events1["start_x"]) < min_dribble_length
    same_y = abs(df_events["end_y"] - df_events1["start_y"]) < min_dribble_length
    same_loc = same_x & same_y

    # df_events.loc[selector_touch_same_player & same_loc, 'subtype_id'] = 70
    # df_events.loc[selector_touch_same_player & same_loc, 'accurate'] = True
    # df_events.loc[selector_touch_same_player & same_loc, 'not_accurate'] = False

    df_events.loc[selector_touch_same_team & same_loc, "type_id"] = 8
    df_events.loc[selector_touch_same_team & same_loc, "subtype_id"] = 85
    df_events.loc[selector_touch_same_team & same_loc, "accurate"] = True
    df_events.loc[selector_touch_same_team & same_loc, "not_accurate"] = False

    df_events.loc[selector_touch_other & same_loc, "type_id"] = 8
    df_events.loc[selector_touch_other & same_loc, "subtype_id"] = 85
    df_events.loc[selector_touch_other & same_loc, "accurate"] = False
    df_events.loc[selector_touch_other & same_loc, "not_accurate"] = True

    return df_events


def create_df_actions(df_events: pd.DataFrame) -> pd.DataFrame:
    """Create the SciSports action dataframe.

    Parameters
    ----------
    df_events : pd.DataFrame
        Wyscout event dataframe

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe
    """
    df_events["time_seconds"] = df_events["milliseconds"] / 1000
    df_actions = df_events[
        [
            "game_id",
            "period_id",
            "time_seconds",
            "team_id",
            "player_id",
            "start_x",
            "start_y",
            "end_x",
            "end_y",
        ]
    ].copy()
    df_actions["original_event_id"] = df_events["event_id"].astype(object)
    df_actions["bodypart_id"] = df_events.apply(determine_bodypart_id, axis=1)
    df_actions["type_id"] = df_events.apply(determine_type_id, axis=1)
    df_actions["result_id"] = df_events.apply(determine_result_id, axis=1)

    df_actions = remove_non_actions(df_actions)  # remove all non-actions left

    return df_actions


def determine_bodypart_id(event: pd.DataFrame) -> int:
    """Determint eht body part for each action.

    Parameters
    ----------
    event : pd.Series
        Wyscout event Series

    Returns
    -------
    int
        id of the body part used for the action
    """
    if event["subtype_id"] in [81, 36, 21, 90, 91]:
        body_part = "other"
    elif event["subtype_id"] == 82:
        body_part = "head"
    elif event["type_id"] == 10 and event["head/body"]:
        body_part = "head/other"
    else:  # all other cases
        body_part = "foot"
    return spadlconfig.bodyparts.index(body_part)


def determine_type_id(event: pd.DataFrame) -> int:  # noqa: C901
    """Determine the type of each action.

    This function transforms the Wyscout events, sub_events and tags
    into the corresponding SciSports action type

    Parameters
    ----------
    event : pd.Series
        A series from the Wyscout event dataframe

    Returns
    -------
    int
        id of the action type
    """
    if event["fairplay"]:
        action_type = "non_action"
    elif event["own_goal"]:
        action_type = "bad_touch"
    elif event["type_id"] == 8:
        if event["subtype_id"] == 80:
            action_type = "cross"
        else:
            action_type = "pass"
    elif event["subtype_id"] == 36:
        action_type = "throw_in"
    elif event["subtype_id"] == 30:
        if event["high"]:
            action_type = "corner_crossed"
        else:
            action_type = "corner_short"
    elif event["subtype_id"] == 32:
        action_type = "freekick_crossed"
    elif event["subtype_id"] == 31:
        action_type = "freekick_short"
    elif event["subtype_id"] == 34:
        action_type = "goalkick"
    elif event["type_id"] == 2 and (event["subtype_id"] not in [22, 23, 24, 26]):
        action_type = "foul"
    elif event["type_id"] == 10:
        action_type = "shot"
    elif event["subtype_id"] == 35:
        action_type = "shot_penalty"
    elif event["subtype_id"] == 33:
        action_type = "shot_freekick"
    elif event["type_id"] == 9:
        action_type = "keeper_save"
    elif event["subtype_id"] == 71:
        action_type = "clearance"
    elif event["subtype_id"] == 72 and event["not_accurate"]:
        action_type = "bad_touch"
    elif event["subtype_id"] == 70:
        action_type = "dribble"
    elif event["take_on_left"] or event["take_on_right"]:
        action_type = "take_on"
    elif event["sliding_tackle"]:
        action_type = "tackle"
    elif event["interception"] and (event["subtype_id"] in [0, 10, 11, 12, 13, 72]):
        action_type = "interception"
    else:
        action_type = "non_action"
    return spadlconfig.actiontypes.index(action_type)


def determine_result_id(event: pd.DataFrame) -> int:  # noqa: C901
    """Determine the result of each event.

    Parameters
    ----------
    event : pd.Series
        Wyscout event Series

    Returns
    -------
    int
        result of the action
    """
    if event["offside"] == 1:
        return 2
    if event["type_id"] == 2:  # foul
        return 1
    if event["goal"]:  # goal
        return 1
    if event["own_goal"]:  # own goal
        return 3
    if event["subtype_id"] in [100, 33, 35]:  # no goal, so 0
        return 0
    if event["accurate"]:
        return 1
    if event["not_accurate"]:
        return 0
    if (
        event["interception"] or event["clearance"] or event["subtype_id"] == 71
    ):  # interception or clearance always success
        return 1
    if event["type_id"] == 9:  # keeper save always success
        return 1
    # no idea, assume it was successful
    return 1


def remove_non_actions(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Remove the remaining non_actions from the action dataframe.

    Parameters
    ----------
    df_actions : pd.DataFrame
        SciSports action dataframe

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe without non-actions
    """
    df_actions = df_actions[df_actions["type_id"] != spadlconfig.actiontypes.index("non_action")]
    # remove remaining ball out of field, whistle and goalkeeper from line
    df_actions = df_actions.reset_index(drop=True)
    return df_actions


def fix_actions(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Fix the generated actions.

    Parameters
    ----------
    df_actions : pd.DataFrame
        SPADL actions dataframe

    Returns
    -------
    pd.DataFrame
        SpADL actions dataframe with end coordinates for shots
    """
    df_actions["start_x"] = (df_actions["start_x"] * spadlconfig.field_length / 100).clip(
        0, spadlconfig.field_length
    )
    df_actions["start_y"] = (
        (100 - df_actions["start_y"])
        * spadlconfig.field_width
        / 100
        # y is from top to bottom in Wyscout
    ).clip(0, spadlconfig.field_width)
    df_actions["end_x"] = (df_actions["end_x"] * spadlconfig.field_length / 100).clip(
        0, spadlconfig.field_length
    )
    df_actions["end_y"] = (
        (100 - df_actions["end_y"])
        * spadlconfig.field_width
        / 100
        # y is from top to bottom in Wyscout
    ).clip(0, spadlconfig.field_width)
    df_actions = fix_goalkick_coordinates(df_actions)
    df_actions = adjust_goalkick_result(df_actions)
    df_actions = fix_foul_coordinates(df_actions)
    df_actions = fix_keeper_save_coordinates(df_actions)
    df_actions = remove_keeper_goal_actions(df_actions)
    df_actions.reset_index(drop=True, inplace=True)

    return df_actions


def fix_goalkick_coordinates(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Fix goalkick coordinates.

    This function sets the goalkick start coordinates to (5,34)

    Parameters
    ----------
    df_actions : pd.DataFrame
        SciSports action dataframe with start coordinates for goalkicks in the
        corner of the pitch

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe including start coordinates for goalkicks
    """
    goalkicks_idx = df_actions["type_id"] == spadlconfig.actiontypes.index("goalkick")
    df_actions.loc[goalkicks_idx, "start_x"] = 5.0
    df_actions.loc[goalkicks_idx, "start_y"] = 34.0

    return df_actions


def fix_foul_coordinates(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Fix fould coordinates.

    This function sets foul end coordinates equal to the foul start coordinates

    Parameters
    ----------
    df_actions : pd.DataFrame
        SciSports action dataframe with no end coordinates for fouls

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe including start coordinates for goalkicks
    """
    fouls_idx = df_actions["type_id"] == spadlconfig.actiontypes.index("foul")
    df_actions.loc[fouls_idx, "end_x"] = df_actions.loc[fouls_idx, "start_x"]
    df_actions.loc[fouls_idx, "end_y"] = df_actions.loc[fouls_idx, "start_y"]

    return df_actions


def fix_keeper_save_coordinates(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Fix keeper save coordinates.

    This function sets keeper_save start coordinates equal to
    keeper_save end coordinates. It also inverts the shot coordinates to the own goal.

    Parameters
    ----------
    df_actions : pd.DataFrame
        SciSports action dataframe with start coordinates in the corner of the pitch

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe with correct keeper_save coordinates
    """
    saves_idx = df_actions["type_id"] == spadlconfig.actiontypes.index("keeper_save")
    # invert the coordinates
    df_actions.loc[saves_idx, "end_x"] = (
        spadlconfig.field_length - df_actions.loc[saves_idx, "end_x"]
    )
    df_actions.loc[saves_idx, "end_y"] = (
        spadlconfig.field_width - df_actions.loc[saves_idx, "end_y"]
    )
    # set start coordinates equal to start coordinates
    df_actions.loc[saves_idx, "start_x"] = df_actions.loc[saves_idx, "end_x"]
    df_actions.loc[saves_idx, "start_y"] = df_actions.loc[saves_idx, "end_y"]

    return df_actions


def remove_keeper_goal_actions(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Remove keeper goal-saving actions.

    This function removes keeper_save actions that appear directly after a goal

    Parameters
    ----------
    df_actions : pd.DataFrame
        SciSports action dataframe with keeper actions directly after a goal

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe without keeper actions directly after a goal
    """
    prev_actions = df_actions.shift(1)
    same_phase = prev_actions.time_seconds + 10 > df_actions.time_seconds
    shot_goals = (prev_actions.type_id == spadlconfig.actiontypes.index("shot")) & (
        prev_actions.result_id == 1
    )
    penalty_goals = (prev_actions.type_id == spadlconfig.actiontypes.index("shot_penalty")) & (
        prev_actions.result_id == 1
    )
    freekick_goals = (prev_actions.type_id == spadlconfig.actiontypes.index("shot_freekick")) & (
        prev_actions.result_id == 1
    )
    goals = shot_goals | penalty_goals | freekick_goals
    keeper_save = df_actions["type_id"] == spadlconfig.actiontypes.index("keeper_save")
    goals_keepers_idx = same_phase & goals & keeper_save
    df_actions = df_actions.drop(df_actions.index[goals_keepers_idx])
    df_actions = df_actions.reset_index(drop=True)

    return df_actions


def adjust_goalkick_result(df_actions: pd.DataFrame) -> pd.DataFrame:
    """Adjust goalkick results.

    This function adjusts goalkick results depending on whether
    the next action is performed by the same team or not

    Parameters
    ----------
    df_actions : pd.DataFrame
        SciSports action dataframe with incorrect goalkick results

    Returns
    -------
    pd.DataFrame
        SciSports action dataframe with correct goalkick results
    """
    nex_actions = df_actions.shift(-1)
    goalkicks = df_actions["type_id"] == spadlconfig.actiontypes.index("goalkick")
    same_team = df_actions["team_id"] == nex_actions["team_id"]
    accurate = same_team & goalkicks
    not_accurate = ~same_team & goalkicks
    df_actions.loc[accurate, "result_id"] = 1
    df_actions.loc[not_accurate, "result_id"] = 0

    return df_actions



"""Utility functions for working with SPADL dataframes."""
from typing import cast

from pandera.typing import DataFrame

from . import config as spadlconfig
from .schema import SPADLSchema


def add_names(actions: DataFrame[SPADLSchema]) -> DataFrame[SPADLSchema]:
    """Add the type name, result name and bodypart name to a SPADL dataframe.

    Parameters
    ----------
    actions : pd.DataFrame
        A SPADL dataframe.

    Returns
    -------
    pd.DataFrame
        The original dataframe with a 'type_name', 'result_name' and
        'bodypart_name' appended.
    """
    return cast(
        DataFrame[SPADLSchema],
        actions.drop(columns=["type_name", "result_name", "bodypart_name"], errors="ignore")
        .merge(spadlconfig.actiontypes_df(), how="left")
        .merge(spadlconfig.results_df(), how="left")
        .merge(spadlconfig.bodyparts_df(), how="left")
        .set_index(actions.index),
    )


def play_left_to_right(
    actions: DataFrame[SPADLSchema], home_team_id: int
) -> DataFrame[SPADLSchema]:
    """Perform all action in the same playing direction.

    This changes the start and end location of each action, such that all actions
    are performed as if the team plays from left to right.

    Parameters
    ----------
    actions : pd.DataFrame
        The SPADL actins of a game.
    home_team_id : int
        The ID of the home team.

    Returns
    -------
    list(pd.DataFrame)
        All actions performed left to right.
    """
    ltr_actions = actions.copy()
    away_idx = actions.team_id != home_team_id
    for col in ["start_x", "end_x"]:
        ltr_actions.loc[away_idx, col] = spadlconfig.field_length - actions[away_idx][col].values
    for col in ["start_y", "end_y"]:
        ltr_actions.loc[away_idx, col] = spadlconfig.field_width - actions[away_idx][col].values
    return ltr_actions



"""Configuration of the Atomic-SPADL language.

Attributes
----------
field_length : float
    The length of a pitch (in meters).
field_width : float
    The width of a pitch (in meters).
bodyparts : list(str)
    The bodyparts used in the Atomic-SPADL language.
actiontypes : list(str)
    The action types used in the Atomic-SPADL language.

"""
import pandas as pd

import socceraction.spadl.config as _spadl

field_length = _spadl.field_length
field_width = _spadl.field_width

bodyparts = _spadl.bodyparts
bodyparts_df = _spadl.bodyparts_df

actiontypes = _spadl.actiontypes + [
    'receival',
    'interception',
    'out',
    'offside',
    'goal',
    'owngoal',
    'yellow_card',
    'red_card',
    'corner',
    'freekick',
]


def actiontypes_df() -> pd.DataFrame:
    """Return a dataframe with the type id and type name of each Atomic-SPADL action type.

    Returns
    -------
    pd.DataFrame
        The 'type_id' and 'type_name' of each Atomic-SPADL action type.
    """
    return pd.DataFrame(list(enumerate(actiontypes)), columns=['type_id', 'type_name'])
